"""Cell class for TensorMesh."""

import itertools
import numpy as np


class TensorCell:
    """
    Representation of a cell in a TensorMesh.

    Parameters
    ----------
    h : (dim) numpy.ndarray
        Array with the cell widths along each direction. For a 2D mesh, it
        must have two elements (``hx``, ``hy``). For a 3D mesh it must have
        three elements (``hx``, ``hy``, ``hz``).
    origin : (dim) numpy.ndarray
        Array with the coordinates of the origin of the cell, i.e. the
        bottom-left-frontmost corner.
    index_unraveled : (dim) tuple
        Array with the unraveled indices of the cell in its parent mesh.
    mesh_shape : (dim) tuple
        Shape of the parent mesh.

    Examples
    --------
    Define a simple :class:`discretize.TensorMesh`.

    >>> from discretize import TensorMesh
    >>> mesh = TensorMesh([5, 8, 10])

    We can obtain a particular cell in the mesh by its index:

    >>> cell = mesh[3]
    >>> cell
    TensorCell(h=[0.2   0.125 0.1  ], origin=[0.6 0.  0. ], index=3, mesh_shape=(5, 8, 10))

    And then obtain information about it, like its
    :attr:`discretize.tensor_cell.TensorCell.origin`:

    >>> cell.origin
    array([0.6, 0. , 0. ])

    Or its
    :attr:`discretize.tensor_cell.TensorCell.bounds`:

    >>> cell.bounds
    array([0.6  , 0.8  , 0.   , 0.125, 0.   , 0.1  ])

    We can also get its neighboring cells:

    >>> neighbours = cell.get_neighbors(mesh)
    >>> for neighbor in neighbours:
    ...     print(neighbor.center)
    [0.5    0.0625 0.05  ]
    [0.9    0.0625 0.05  ]
    [0.7    0.1875 0.05  ]
    [0.7    0.0625 0.15  ]


    Alternatively, we can iterate over all cells in the mesh with a simple
    *for loop* or list comprehension:

    >>> cells = [cell for cell in mesh]
    >>> len(cells)
    400

    """

    def __init__(self, h, origin, index_unraveled, mesh_shape):
        self._h = h
        self._origin = origin
        self._index_unraveled = index_unraveled
        self._mesh_shape = mesh_shape

    def __repr__(self):
        """Represent a TensorCell."""
        attributes = ", ".join(
            [
                f"{attr}={getattr(self, attr)}"
                for attr in ("h", "origin", "index", "mesh_shape")
            ]
        )
        return f"TensorCell({attributes})"

    def __eq__(self, other):
        """Check if this cell is the same as other one."""
        if not isinstance(other, TensorCell):
            raise TypeError(
                f"Cannot compare an object of type '{other.__class__.__name__}' "
                "with a TensorCell"
            )
        are_equal = (
            np.all(self.h == other.h)
            and np.all(self.origin == other.origin)
            and self.index == other.index
            and self.mesh_shape == other.mesh_shape
        )
        return are_equal

    @property
    def h(self):
        """Cell widths."""
        return self._h

    @property
    def origin(self):
        """Coordinates of the origin of the cell."""
        return self._origin

    @property
    def index(self):
        """Index of the cell in a TensorMesh."""
        return np.ravel_multi_index(
            self.index_unraveled, dims=self.mesh_shape, order="F"
        )

    @property
    def index_unraveled(self):
        """Unraveled index of the cell in a TensorMesh."""
        return self._index_unraveled

    @property
    def mesh_shape(self):
        """Shape of the parent mesh."""
        return self._mesh_shape

    @property
    def dim(self):
        """Dimensions of the cell (1, 2 or 3)."""
        return len(self.h)

    @property
    def center(self):
        """
        Coordinates of the cell center.

        Returns
        -------
        center : (dim) array
            Array with the coordinates of the cell center.
        """
        center = np.array(self.origin) + np.array(self.h) / 2
        return center

    @property
    def bounds(self):
        """
        Bounds of the cell.

        Coordinates that define the bounds of the cell. Bounds are returned in
        the following order: ``x1``, ``x2``, ``y1``, ``y2``, ``z1``, ``z2``.

        Returns
        -------
        bounds : (2 * dim) array
            Array with the cell bounds.
        """
        bounds = np.array(
            [
                origin_i + factor * h_i
                for origin_i, h_i in zip(self.origin, self.h)
                for factor in (0, 1)
            ]
        )
        return bounds

    @property
    def neighbors(self):
        """
        Indices for this cell's neighbors within its parent mesh.

        Returns
        -------
        list of list of int
        """
        neighbor_indices = []
        for dim in range(self.dim):
            for delta in (-1, 1):
                index = list(self.index_unraveled)
                index[dim] += delta
                if 0 <= index[dim] < self._mesh_shape[dim]:
                    neighbor_indices.append(
                        np.ravel_multi_index(index, dims=self.mesh_shape, order="F")
                    )
        return neighbor_indices

    @property
    def nodes(self):
        """
        Indices for this cell's nodes within its parent mesh.

        Returns
        -------
        list of int
        """
        # Define shape of nodes in parent mesh
        nodes_shape = [s + 1 for s in self.mesh_shape]
        # Get indices of nodes per dimension
        nodes_index_per_dim = [[index, index + 1] for index in self.index_unraveled]
        # Combine the nodes_index_per_dim using itertools.product.
        # Because we want to follow a FORTRAN order, we need to reverse the
        # order of the nodes_index_per_dim and the indices.
        nodes_indices = [i[::-1] for i in itertools.product(*nodes_index_per_dim[::-1])]
        # Ravel indices
        nodes_indices = [
            np.ravel_multi_index(index, dims=nodes_shape, order="F")
            for index in nodes_indices
        ]
        return nodes_indices

    @property
    def edges(self):
        """
        Indices for this cell's edges within its parent mesh.

        Returns
        -------
        list of int
        """
        if self.dim == 1:
            edges_indices = [self.index]
        elif self.dim == 2:
            # Get shape of edges grids (for edges_x and edges_y)
            edges_x_shape = [self.mesh_shape[0], self.mesh_shape[1] + 1]
            edges_y_shape = [self.mesh_shape[0] + 1, self.mesh_shape[1]]
            # Calculate total amount of edges_x
            n_edges_x = edges_x_shape[0] * edges_x_shape[1]
            # Get indices of edges_x
            edges_x_indices = [
                [self.index_unraveled[0], self.index_unraveled[1] + delta]
                for delta in (0, 1)
            ]
            edges_x_indices = [
                np.ravel_multi_index(index, dims=edges_x_shape, order="F")
                for index in edges_x_indices
            ]
            # Get indices of edges_y
            edges_y_indices = [
                [self.index_unraveled[0] + delta, self.index_unraveled[1]]
                for delta in (0, 1)
            ]
            edges_y_indices = [
                n_edges_x + np.ravel_multi_index(index, dims=edges_y_shape, order="F")
                for index in edges_y_indices
            ]
            edges_indices = edges_x_indices + edges_y_indices
        elif self.dim == 3:
            edges_x_shape = [
                n if i == 0 else n + 1 for i, n in enumerate(self.mesh_shape)
            ]
            edges_y_shape = [
                n if i == 1 else n + 1 for i, n in enumerate(self.mesh_shape)
            ]
            edges_z_shape = [
                n if i == 2 else n + 1 for i, n in enumerate(self.mesh_shape)
            ]
            # Calculate total amount of edges_x and edges_y
            n_edges_x = edges_x_shape[0] * edges_x_shape[1] * edges_x_shape[2]
            n_edges_y = edges_y_shape[0] * edges_y_shape[1] * edges_y_shape[2]
            # Get indices of edges_x
            edges_x_indices = [
                [
                    self.index_unraveled[0],
                    self.index_unraveled[1] + delta_y,
                    self.index_unraveled[2] + delta_z,
                ]
                for delta_z in (0, 1)
                for delta_y in (0, 1)
            ]
            edges_x_indices = [
                np.ravel_multi_index(index, dims=edges_x_shape, order="F")
                for index in edges_x_indices
            ]
            # Get indices of edges_y
            edges_y_indices = [
                [
                    self.index_unraveled[0] + delta_x,
                    self.index_unraveled[1],
                    self.index_unraveled[2] + delta_z,
                ]
                for delta_z in (0, 1)
                for delta_x in (0, 1)
            ]
            edges_y_indices = [
                n_edges_x + np.ravel_multi_index(index, dims=edges_y_shape, order="F")
                for index in edges_y_indices
            ]
            # Get indices of edges_z
            edges_z_indices = [
                [
                    self.index_unraveled[0] + delta_x,
                    self.index_unraveled[1] + delta_y,
                    self.index_unraveled[2],
                ]
                for delta_y in (0, 1)
                for delta_x in (0, 1)
            ]
            edges_z_indices = [
                n_edges_x
                + n_edges_y
                + np.ravel_multi_index(index, dims=edges_z_shape, order="F")
                for index in edges_z_indices
            ]
            edges_indices = edges_x_indices + edges_y_indices + edges_z_indices
        return edges_indices

    @property
    def faces(self):
        """
        Indices for cell's faces within its parent mesh.

        Returns
        -------
        list of int
        """
        if self.dim == 1:
            faces_indices = [self.index, self.index + 1]
        elif self.dim == 2:
            # Get shape of faces grids
            # (faces_x are normal to x and faces_y are normal to y)
            faces_x_shape = [self.mesh_shape[0] + 1, self.mesh_shape[1]]
            faces_y_shape = [self.mesh_shape[0], self.mesh_shape[1] + 1]
            # Calculate total amount of faces_x
            n_faces_x = faces_x_shape[0] * faces_x_shape[1]
            # Get indices of faces_x
            faces_x_indices = [
                [self.index_unraveled[0] + delta, self.index_unraveled[1]]
                for delta in (0, 1)
            ]
            faces_x_indices = [
                np.ravel_multi_index(index, dims=faces_x_shape, order="F")
                for index in faces_x_indices
            ]
            # Get indices of faces_y
            faces_y_indices = [
                [self.index_unraveled[0], self.index_unraveled[1] + delta]
                for delta in (0, 1)
            ]
            faces_y_indices = [
                n_faces_x + np.ravel_multi_index(index, dims=faces_y_shape, order="F")
                for index in faces_y_indices
            ]
            faces_indices = faces_x_indices + faces_y_indices
        elif self.dim == 3:
            # Get shape of faces grids
            faces_x_shape = [
                n + 1 if i == 0 else n for i, n in enumerate(self.mesh_shape)
            ]
            faces_y_shape = [
                n + 1 if i == 1 else n for i, n in enumerate(self.mesh_shape)
            ]
            faces_z_shape = [
                n + 1 if i == 2 else n for i, n in enumerate(self.mesh_shape)
            ]
            # Calculate total amount of faces_x and faces_y
            n_faces_x = faces_x_shape[0] * faces_x_shape[1] * faces_x_shape[2]
            n_faces_y = faces_y_shape[0] * faces_y_shape[1] * faces_y_shape[2]
            # Get indices of faces_x
            faces_x_indices = [
                [
                    self.index_unraveled[0] + delta,
                    self.index_unraveled[1],
                    self.index_unraveled[2],
                ]
                for delta in (0, 1)
            ]
            faces_x_indices = [
                np.ravel_multi_index(index, dims=faces_x_shape, order="F")
                for index in faces_x_indices
            ]
            # Get indices of faces_y
            faces_y_indices = [
                [
                    self.index_unraveled[0],
                    self.index_unraveled[1] + delta,
                    self.index_unraveled[2],
                ]
                for delta in (0, 1)
            ]
            faces_y_indices = [
                n_faces_x + np.ravel_multi_index(index, dims=faces_y_shape, order="F")
                for index in faces_y_indices
            ]
            # Get indices of faces_z
            faces_z_indices = [
                [
                    self.index_unraveled[0],
                    self.index_unraveled[1],
                    self.index_unraveled[2] + delta,
                ]
                for delta in (0, 1)
            ]
            faces_z_indices = [
                n_faces_x
                + n_faces_y
                + np.ravel_multi_index(index, dims=faces_z_shape, order="F")
                for index in faces_z_indices
            ]
            faces_indices = faces_x_indices + faces_y_indices + faces_z_indices
        return faces_indices

    def get_neighbors(self, mesh):
        """
        Return the neighboring cells in the mesh.

        Parameters
        ----------
        mesh : TensorMesh
            TensorMesh where the current cell lives.

        Returns
        -------
        list of TensorCell
        """
        return [mesh[index] for index in self.neighbors]



"""
===========================================
Testing Utilities (:mod:`discretize.tests`)
===========================================
.. currentmodule:: discretize.tests

This module contains utilities for convergence testing.

Classes
-------
.. autosummary::
  :toctree: generated/

  OrderTest

Functions
---------
.. autosummary::
  :toctree: generated/

  check_derivative
  rosenbrock
  get_quadratic
  setup_mesh
  assert_isadjoint
"""  # NOQA D205

import numpy as np
import scipy.sparse as sp

from discretize.utils import mkvc, example_curvilinear_grid, requires
from discretize.tensor_mesh import TensorMesh
from discretize.curvilinear_mesh import CurvilinearMesh
from discretize.cylindrical_mesh import CylindricalMesh
from discretize.utils.code_utils import deprecate_function

from . import TreeMesh as Tree

import unittest
import inspect

try:
    import getpass

    name = getpass.getuser()[0].upper() + getpass.getuser()[1:]
except Exception:
    name = "You"

happiness = [
    "The test be workin!",
    "You get a gold star!",
    "Yay passed!",
    "Happy little convergence test!",
    "That was easy!",
    "Testing is important.",
    "You are awesome.",
    "Go Test Go!",
    "Once upon a time, a happy little test passed.",
    "And then everyone was happy.",
    "Not just a pretty face " + name,
    "You deserve a pat on the back!",
    "Well done " + name + "!",
    "Awesome, " + name + ", just awesome.",
]
sadness = [
    "No gold star for you.",
    "Try again soon.",
    "Thankfully,  persistence is a great substitute for talent.",
    "It might be easier to call this a feature...",
    "Coffee break?",
    "Boooooooo  :(",
    "Testing is important. Do it again.",
    "Did you put your clever trousers on today?",
    "Just think about a dancing dinosaur and life will get better!",
    "You had so much promise " + name + ", oh well...",
    name.upper() + " ERROR!",
    "Get on it " + name + "!",
    "You break it, you fix it.",
]

_happiness_rng = np.random.default_rng()


def setup_mesh(mesh_type, nC, nDim, random_seed=None):
    """Generate arbitrary mesh for testing.

    For the mesh type, number of cells along each axis and dimension specified,
    **setup_mesh** will construct a random mesh that can be used for testing.
    By design, the domain width is 1 along each axis direction.

    Parameters
    ----------
    mesh_type : str
        Defines the mesh type. Must be one of **{'uniformTensorMesh',
        'randomTensorMesh', 'uniformCylindricalMesh', 'randomCylindricalMesh',
        'uniformTree', 'randomTree', 'uniformCurv', 'rotateCurv', 'sphereCurv'}**
    nC : int
        Number of cells along each axis. If *mesh_type* is 'Tree', then *nC* defines the
        number of base mesh cells and must be a power of 2.
    nDim : int
        The dimension of the mesh. Must be 1, 2 or 3.
    random_seed : numpy.random.Generator, int, optional
        If ``random`` is in `mesh_type`, this is the random number generator to use for
        creating that random mesh. If an integer or None it is used to seed a new
        `numpy.random.default_rng`.

    Returns
    -------
    discretize.base.BaseMesh
        A discretize mesh of class specified by the input argument *mesh_type*
    """
    if "random" in mesh_type:
        rng = np.random.default_rng(random_seed)
    if "TensorMesh" in mesh_type:
        if "uniform" in mesh_type:
            h = [nC, nC, nC]
        elif "random" in mesh_type:
            h1 = rng.random(nC) * nC * 0.5 + nC * 0.5
            h2 = rng.random(nC) * nC * 0.5 + nC * 0.5
            h3 = rng.random(nC) * nC * 0.5 + nC * 0.5
            h = [hi / np.sum(hi) for hi in [h1, h2, h3]]  # normalize
        else:
            raise Exception("Unexpected mesh_type")

        mesh = TensorMesh(h[:nDim])
        max_h = max([np.max(hi) for hi in mesh.h])

    elif "CylindricalMesh" in mesh_type or "CylMesh" in mesh_type:
        if "uniform" in mesh_type:
            if "symmetric" in mesh_type:
                h = [nC, 1, nC]
            else:
                h = [nC, nC, nC]
        elif "random" in mesh_type:
            h1 = rng.random(nC) * nC * 0.5 + nC * 0.5
            if "symmetric" in mesh_type:
                h2 = [
                    2 * np.pi,
                ]
            else:
                h2 = rng.random(nC) * nC * 0.5 + nC * 0.5
            h3 = rng.random(nC) * nC * 0.5 + nC * 0.5
            h = [hi / np.sum(hi) for hi in [h1, h2, h3]]  # normalize
            h[1] = h[1] * 2 * np.pi
        else:
            raise Exception("Unexpected mesh_type")

        if nDim == 2:
            mesh = CylindricalMesh([h[0], h[1]])
            if "symmetric" in mesh_type:
                max_h = np.max(mesh.h[0])
            else:
                max_h = max([np.max(hi) for hi in mesh.h])
        elif nDim == 3:
            mesh = CylindricalMesh(h)
            if "symmetric" in mesh_type:
                max_h = max([np.max(hi) for hi in [mesh.h[0], mesh.h[2]]])
            else:
                max_h = max([np.max(hi) for hi in mesh.h])

    elif "Curv" in mesh_type:
        if "uniform" in mesh_type:
            kwrd = "rect"
        elif "rotate" in mesh_type:
            kwrd = "rotate"
        elif "sphere" in mesh_type:
            kwrd = "sphere"
        else:
            raise Exception("Unexpected mesh_type")
        if nDim == 1:
            raise Exception("Lom not supported for 1D")
        elif nDim == 2:
            X, Y = example_curvilinear_grid([nC, nC], kwrd)
            mesh = CurvilinearMesh([X, Y])
        elif nDim == 3:
            X, Y, Z = example_curvilinear_grid([nC, nC, nC], kwrd)
            mesh = CurvilinearMesh([X, Y, Z])
        max_h = 1.0 / nC

    elif "Tree" in mesh_type:
        if Tree is None:
            raise Exception("Tree Mesh not installed. Run 'python setup.py install'")
        nC *= 2
        if "uniform" in mesh_type or "notatree" in mesh_type:
            h = [nC, nC, nC]
        elif "random" in mesh_type:
            h1 = rng.random(nC) * nC * 0.5 + nC * 0.5
            h2 = rng.random(nC) * nC * 0.5 + nC * 0.5
            h3 = rng.random(nC) * nC * 0.5 + nC * 0.5
            h = [hi / np.sum(hi) for hi in [h1, h2, h3]]  # normalize
        else:
            raise Exception("Unexpected mesh_type")

        levels = int(np.log(nC) / np.log(2))
        mesh = Tree(h[:nDim], levels=levels)

        def function(cell):
            if "notatree" in mesh_type:
                return levels - 1
            r = cell.center - 0.5
            dist = np.sqrt(r.dot(r))
            if dist < 0.2:
                return levels
            return levels - 1

        mesh.refine(function)
        # mesh.number()
        # mesh.plot_grid(show_it=True)
        max_h = max([np.max(hi) for hi in mesh.h])
    return mesh, max_h


class OrderTest(unittest.TestCase):
    r"""Base class for testing convergence of discrete operators with respect to cell size.

    ``OrderTest`` is a base class for testing the order of convergence of discrete
    operators with respect to cell size. ``OrderTest`` is inherited by the test
    class for the given operator. Within the test class, the user sets the parameters
    for the convergence testing and defines a method :py:attr:`~OrderTest.getError`
    which defines the error as a norm of the residual (see example).

    OrderTest inherits from :class:`unittest.TestCase`.

    Attributes
    ----------
    name : str
        Name the convergence test
    meshTypes : list of str
        List denoting the mesh types on which the convergence will be tested.
        List entries must be of the list {'uniformTensorMesh', 'randomTensorMesh',
        'uniformCylindricalMesh', 'randomCylindricalMesh', 'uniformTree', 'randomTree',
        'uniformCurv', 'rotateCurv', 'sphereCurv'}
    expectedOrders : float or list of float (default = 2.0)
        Defines the expect orders of convergence for all meshes defined in *meshTypes*.
        If list, must have same length as argument *meshTypes*.
    tolerance : float or list of float (default = 0.85)
        Defines tolerance for numerical approximate of convergence order.
        If list, must have same length as argument *meshTypes*.
    meshSizes : list of int
        From coarsest to finest, defines the number of mesh cells in each axis direction
        for the meshes used in the convergence test; e.g. [4, 8, 16, 32]
    meshDimension : int
        Mesh dimension. Must be 1, 2 or 3
    random_seed : numpy.random.Generator, int, optional
        If ``random`` is in `mesh_type`, this is the random number generator
        used generate the random meshes, if an ``int`` or ``None``, it used to seed
        a new `numpy.random.default_rng`.

    Notes
    -----
    Consider an operator :math:`A(f)` that acts on a test function :math:`f`. And let
    :math:`A_h (f)` be the discrete approximation to the original operator constructed
    on a mesh will cell size :math:`h`. ``OrderTest`` assesses the convergence of

    .. math::
        error(h) = \| A_h(f) - A(f) \|

    as :math:`h \rightarrow 0`. Note that you can provide any norm to quantify the error.
    The convergence test is passed when the numerically estimated rate of convergence is within
    a specified tolerance of the expected convergence rate supplied by the user.

    Examples
    --------
    Here, we utilize the ``OrderTest`` class to validate the rate of convergence for
    the :py:attr:`~discretize.differential_operators.face_divergence`. Our convergence
    test is done for a uniform 2D tensor mesh. Under the test class *TestDIV2D*, we
    define the static parameters for the order test. We then define a method *getError*
    for this class which returns the norm of some residual. With these two pieces
    defined, we can call the order test as shown below.

    >>> from discretize.tests import OrderTest
    >>> import unittest
    >>> import numpy as np

    >>> class TestDIV2D(OrderTest):
    ...     # Static properties for OrderTest
    ...     name = "Face Divergence 2D"
    ...     meshTypes = ["uniformTensorMesh"]
    ...     meshDimension = 2
    ...     expectedOrders = 2.0
    ...     tolerance = 0.85
    ...     meshSizes = [8, 16, 32, 64]
    ...     def getError(self):
    ...         # Test function
    ...         fx = lambda x, y: np.sin(2 * np.pi * x)
    ...         fy = lambda x, y: np.sin(2 * np.pi * y)
    ...         # Analytic solution for operator acting on test function
    ...         sol = lambda x, y: 2 * np.pi * (np.cos(2 * np.pi * x) + np.cos(2 * np.pi * y))
    ...         # Evaluate test function on faces
    ...         f = np.r_[
    ...             fx(self.M.faces_x[:, 0], self.M.faces_x[:, 1]),
    ...             fy(self.M.faces_y[:, 0], self.M.faces_y[:, 1])
    ...         ]
    ...         # Analytic solution at cell centers
    ...         div_f = sol(self.M.cell_centers[:, 0], self.M.cell_centers[:, 1])
    ...         # Numerical approximation of divergence at cell centers
    ...         div_f_num = self.M.face_divergence * f
    ...         # Define the error function as a norm
    ...         err = np.linalg.norm((div_f_num - div_f), np.inf)
    ...         return err
    ...     def test_order(self):
    ...         self.orderTest()
    """

    name = "Order Test"
    expectedOrders = (
        2.0  # This can be a list of orders, must be the same length as meshTypes
    )
    tolerance = 0.85  # This can also be a list, must be the same length as meshTypes
    meshSizes = [4, 8, 16, 32]
    meshTypes = ["uniformTensorMesh"]
    _meshType = meshTypes[0]
    meshDimension = 3
    random_seed = None

    def setupMesh(self, nC):
        """Generate mesh and set as current mesh for testing.

        Parameters
        ----------
        nC : int
            Number of cells along each axis.

        Returns
        -------
        Float
            Maximum cell width for the mesh
        """
        mesh, max_h = setup_mesh(
            self._meshType, nC, self.meshDimension, random_seed=self.random_seed
        )
        self.M = mesh
        return max_h

    def getError(self):
        r"""Compute error defined as a norm of the residual.

        This method is overwritten within the test class of a particular operator.
        Within the method, we define a test function :math:`f`, the analytic solution
        of an operator :math:`A(f)` acting on the test function, and the numerical
        approximation obtained by applying the discretized operator :math:`A_h (f)`.
        **getError** is defined to return the norm of the residual as shown below:

        .. math::
            error(h) = \| A_h(f) - A(f) \|

        """
        return 1.0

    def orderTest(self, random_seed=None):
        """Perform an order test.

        For number of cells specified in meshSizes setup mesh, call getError
        and prints mesh size, error, ratio between current and previous error,
        and estimated order of convergence.
        """
        __tracebackhide__ = True
        if not isinstance(self.meshTypes, list):
            raise TypeError("meshTypes must be a list")
        if type(self.tolerance) is not list:
            self.tolerance = np.ones(len(self.meshTypes)) * self.tolerance

        # if we just provide one expected order, repeat it for each mesh type
        if isinstance(self.expectedOrders, (float, int)):
            self.expectedOrders = [self.expectedOrders for i in self.meshTypes]
        try:
            self.expectedOrders = list(self.expectedOrders)
        except TypeError:
            raise TypeError("expectedOrders must be array like")
        if len(self.expectedOrders) != len(self.meshTypes):
            raise ValueError(
                "expectedOrders must have the same length as the meshTypes"
            )

        if random_seed is not None:
            self.random_seed = random_seed

        def test_func(n_cells):
            max_h = self.setupMesh(n_cells)
            err = self.getError()
            return err, max_h

        for mesh_type, order, tolerance in zip(
            self.meshTypes, self.tolerance, self.expectedOrders
        ):
            self._meshType = mesh_type
            assert_expected_order(
                test_func,
                self.meshSizes,
                expected_order=order,
                rtol=np.abs(1 - tolerance),
                test_type="mean_at_least",
            )


def assert_expected_order(
    func, n_cells, expected_order=2.0, rtol=0.15, test_type="mean"
):
    """Perform an order test.

    For number of cells specified in `mesh_sizes` call `func`
    and prints mesh size, error, ratio between current and previous error,
    and estimated order of convergence.

    Parameters
    ----------
    func : callable
        Function which should accept an integer representing the number of
        discretizations on the domain and return a tuple of the error and
        the discretization widths.
    n_cells : array_like of int
        List of number of discretizations to pass to func.
    expected_order : float, optional
        The expected order of accuracy for you test
    rtol : float, optional
        The relative tolerance of the order test.
    test_type : {'mean', 'min', 'last', 'all', 'mean_at_least'}
        Which property of the list of calculated orders to test.

    Returns
    -------
    numpy.ndarray
        The calculated order values on success

    Raises
    ------
    AssertionError

    Notes
    -----
    For the different ``test_type`` arguments, different properties of the
    order is tested:

        - `mean`: the mean value of all calculated orders is tested for
          approximate equality with the expected order.
        - `min`: The minimimum value of calculated orders is tested for
          approximate equality with the expected order.
        - `last`: The last calculated order is tested for approximate equality
          with the expected order.
        - `all`: All calculated orders are tested for approximate equality with
          the expected order.
        - `mean_at_least`: The mean is tested to be at least approximately the
          expected order. This is the default test for the previous ``OrderTest``
          class in older versions of `discretize`.

    Examples
    --------
    Testing the convergence order of an central difference operator

    >>> from discretize.tests import assert_expected_order
    >>> func = lambda y: np.cos(y)
    >>> func_deriv = lambda y: -np.sin(y)

    Define the function that returns the error and cell width for
    a given number of discretizations.
    >>> def deriv_error(n):
    ...     # grid points
    ...     nodes = np.linspace(0, 1, n+1)
    ...     cc = 0.5 * (nodes[1:] + nodes[:-1])
    ...     dh = nodes[1]-nodes[0]
    ...     # evaluate the function on nodes
    ...     node_eval = func(nodes)
    ...     # calculate the numerical derivative
    ...     num_deriv = (node_eval[1:] - node_eval[:-1]) / dh
    ...     # calculate the true derivative
    ...     true_deriv = func_deriv(cc)
    ...     # compare the L-inf norm of the error vector
    ...     err = np.linalg.norm(num_deriv - true_deriv, ord=np.inf)
    ...     return err, dh

    Then run the expected order test.
    >>> assert_expected_order(deriv_error, [10, 20, 30, 40, 50])
    """
    __tracebackhide__ = True
    n_cells = np.asarray(n_cells, dtype=int)
    if test_type not in ["mean", "min", "last", "all", "mean_at_least"]:
        raise ValueError
    orders = []
    # Do first values:
    nc = n_cells[0]
    err_last, h_last = func(nc)

    print("_______________________________________________________")
    print("  nc  |    h    |    error    | e(i-1)/e(i) |  order   ")
    print("~~~~~~|~~~~~~~~~|~~~~~~~~~~~~~|~~~~~~~~~~~~~|~~~~~~~~~~")
    print(f"{nc:^6d}|{h_last:^9.2e}|{err_last:^13.3e}|             |")

    for nc in n_cells[1:]:
        err, h = func(nc)
        order = np.log(err / err_last) / np.log(h / h_last)
        print(f"{nc:^6d}|{h:^9.2e}|{err:^13.3e}|{err_last / err:^13.4f}|{order:^10.4f}")
        err_last = err
        h_last = h
        orders.append(order)

    print("-------------------------------------------------------")

    try:
        if test_type == "mean":
            np.testing.assert_allclose(np.mean(orders), expected_order, rtol=rtol)
        elif test_type == "mean_at_least":
            test = np.mean(orders) > expected_order * (1 - rtol)
            if not test:
                raise AssertionError(
                    f"\nOrder mean {np.mean(orders)} is not greater than the expected order "
                    f"{expected_order} within the tolerance {rtol}."
                )
        elif test_type == "min":
            np.testing.assert_allclose(np.min(orders), expected_order, rtol=rtol)
        elif test_type == "last":
            np.testing.assert_allclose(orders[-1], expected_order, rtol=rtol)
        elif test_type == "all":
            np.testing.assert_allclose(orders, expected_order, rtol=rtol)
        print(_happiness_rng.choice(happiness))
    except AssertionError as err:
        print(_happiness_rng.choice(sadness))
        raise err

    return orders


def rosenbrock(x, return_g=True, return_H=True):
    """Evaluate the Rosenbrock function.

    This is mostly used for testing Gauss-Newton schemes

    Parameters
    ----------
    x : numpy.ndarray
        The (x0, x1) location for the Rosenbrock test
    return_g : bool, optional
        If *True*, return the gradient at *x*
    return_H : bool, optional
        If *True*, return the approximate Hessian at *x*

    Returns
    -------
    tuple
        Rosenbrock function evaluated at (x0, x1), the gradient at (x0, x1) if
        *return_g = True* and the Hessian at (x0, x1) if *return_H = True*
    """
    f = 100 * (x[1] - x[0] ** 2) ** 2 + (1 - x[0]) ** 2
    g = np.array(
        [2 * (200 * x[0] ** 3 - 200 * x[0] * x[1] + x[0] - 1), 200 * (x[1] - x[0] ** 2)]
    )
    H = sp.csr_matrix(
        np.array(
            [[-400 * x[1] + 1200 * x[0] ** 2 + 2, -400 * x[0]], [-400 * x[0], 200]]
        )
    )

    out = (f,)
    if return_g:
        out += (g,)
    if return_H:
        out += (H,)
    return out if len(out) > 1 else out[0]


def check_derivative(
    fctn,
    x0,
    num=7,
    plotIt=True,
    dx=None,
    expectedOrder=2,
    tolerance=0.85,
    eps=1e-10,
    ax=None,
    random_seed=None,
):
    """Perform a basic derivative check.

    Compares error decay of 0th and 1st order Taylor approximation at point
    x0 for a randomized search direction.

    Parameters
    ----------
    fctn : callable
        The function to test.
    x0 : numpy.ndarray
        Point at which to check derivative
    num : int, optional
        number of times to reduce step length to evaluate derivative
    plotIt : bool, optional
        If *True*, plot the convergence of the approximation of the derivative
    dx : numpy.ndarray, optional
        Step direction. By default, this parameter is set to *None* and a random
        step direction is chosen using `rng`.
    expectedOrder : int, optional
        The expected order of convergence for the numerical derivative
    tolerance : float, optional
        The tolerance on the expected order
    eps : float, optional
        A threshold value for approximately equal to zero
    ax : matplotlib.pyplot.Axes, optional
        An axis object for the convergence plot if *plotIt = True*.
        Otherwise, the function will create a new axis.
    random_seed : numpy.random.Generator, int, optional
        If `dx` is ``None``, this is the random number generator to use for
        generating a step direction. If an integer or None, it is used to seed
        a new `numpy.random.default_rng`.

    Returns
    -------
    bool
        Whether you passed the test.

    Examples
    --------
    >>> from discretize import tests, utils
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> rng = np.random.default_rng(786412)

    >>> def simplePass(x):
    ...     return np.sin(x), utils.sdiag(np.cos(x))
    >>> passed = tests.check_derivative(simplePass, rng.standard_normal(5), random_seed=rng)
    ==================== check_derivative ====================
    iter    h         |ft-f0|   |ft-f0-h*J0*dx|  Order
    ---------------------------------------------------------
     0   1.00e-01    1.690e-01     8.400e-03      nan
     1   1.00e-02    1.636e-02     8.703e-05      1.985
     2   1.00e-03    1.630e-03     8.732e-07      1.999
     3   1.00e-04    1.629e-04     8.735e-09      2.000
     4   1.00e-05    1.629e-05     8.736e-11      2.000
     5   1.00e-06    1.629e-06     8.736e-13      2.000
     6   1.00e-07    1.629e-07     8.822e-15      1.996
    ========================= PASS! =========================
    Once upon a time, a happy little test passed.
    """
    __tracebackhide__ = True
    # matplotlib is a soft dependencies for discretize,
    # lazy-loaded to decrease load time of discretize.

    try:
        import matplotlib
        import matplotlib.pyplot as plt
    except ImportError:
        matplotlib = False

    print("{0!s} check_derivative {1!s}".format("=" * 20, "=" * 20))
    print(
        "iter    h         |ft-f0|   |ft-f0-h*J0*dx|  Order\n{0!s}".format(("-" * 57))
    )

    f0, J0 = fctn(x0)

    x0 = mkvc(x0)

    if dx is None:
        rng = np.random.default_rng(random_seed)
        dx = rng.standard_normal(len(x0))

    h = np.logspace(-1, -num, num)
    E0 = np.ones(h.shape)
    E1 = np.ones(h.shape)

    def l2norm(x):
        # because np.norm breaks if they are scalars?
        return np.sqrt(np.real(np.vdot(x, x)))

    for i in range(num):
        # Evaluate at test point
        ft, Jt = fctn(x0 + h[i] * dx)
        # 0th order Taylor
        E0[i] = l2norm(ft - f0)
        # 1st order Taylor
        if inspect.isfunction(J0):
            E1[i] = l2norm(ft - f0 - h[i] * J0(dx))
        else:
            # We assume it is a numpy.ndarray
            E1[i] = l2norm(ft - f0 - h[i] * J0.dot(dx))

        order0 = np.log10(E0[:-1] / E0[1:])
        order1 = np.log10(E1[:-1] / E1[1:])
        print(
            " {0:d}   {1:1.2e}    {2:1.3e}     {3:1.3e}      {4:1.3f}".format(
                i, h[i], E0[i], E1[i], np.nan if i == 0 else order1[i - 1]
            )
        )

    @requires({"matplotlib": matplotlib})
    def _plot_it(axes, passed):
        if axes is None:
            axes = plt.subplot(111)
        axes.loglog(h, E0, "b")
        axes.loglog(h, E1, "g--")
        axes.set_title(
            "Check Derivative - {0!s}".format(("PASSED :)" if passed else "FAILED :("))
        )
        axes.set_xlabel("h")
        axes.set_ylabel("Error")
        leg = axes.legend(
            [r"$\mathcal{O}(h)$", r"$\mathcal{O}(h^2)$"],
            loc="best",
            title=r"$f(x + h\Delta x) - f(x) - h g(x) \Delta x - \mathcal{O}(h^2) = 0$",
            frameon=False,
        )
        plt.setp(leg.get_title(), fontsize=15)
        plt.show()

    # Ensure we are about precision
    order0 = order0[E0[1:] > eps]
    order1 = order1[E1[1:] > eps]

    # belowTol = order1.size == 0 and order0.size >= 0
    # # Make sure we get the correct order
    # correctOrder = order1.size > 0 and np.mean(order1) > tolerance * expectedOrder
    #
    # passTest = belowTol or correctOrder
    try:
        if order1.size == 0:
            # This should happen if all of the 1st order taylor approximation errors
            # were below epsilon, common if the original function was linear.
            # Thus it has no higher order derivatives.
            pass
        else:
            order_mean = np.mean(order1)
            expected = tolerance * expectedOrder
            test = order_mean > expected
            if not test:
                raise AssertionError(
                    f"\n Order mean {order_mean} is not greater than"
                    f" {expected} = tolerance: {tolerance} "
                    f"* expected order: {expectedOrder}."
                )
        print("{0!s} PASS! {1!s}".format("=" * 25, "=" * 25))
        print(_happiness_rng.choice(happiness) + "\n")
        if plotIt:
            _plot_it(ax, True)
    except AssertionError as err:
        print(
            "{0!s}\n{1!s} FAIL! {2!s}\n{3!s}".format(
                "*" * 57, "<" * 25, ">" * 25, "*" * 57
            )
        )
        print(_happiness_rng.choice(sadness) + "\n")
        if plotIt:
            _plot_it(ax, False)
        raise err

    return True


def get_quadratic(A, b, c=0):
    r"""Return a function that evaluates the given quadratic.

    Given **A**, **b** and *c*, this returns a function that evaluates
    the quadratic for a vector **x**. Where :math:`\mathbf{A} \in \mathbb{R}^{NxN}`,
    :math:`\mathbf{b} \in \mathbb{R}^N` and :math:`c` is a constant,
    this function evaluates the following quadratic:

    .. math::

        Q( \mathbf{x} ) = \frac{1}{2} \mathbf{x^T A x + b^T x} + c

    for a vector :math:`\mathbf{x}`. It also optionally returns the gradient of the
    above equation, and its Hessian.

    Parameters
    ----------
    A : (N, N) numpy.ndarray
        A square matrix
    b : (N) numpy.ndarray
        A vector
    c : float
        A constant

    Returns
    -------
    function :
        The callable function that returns the quadratic evaluation, and optionally its
        gradient, and Hessian.
    """

    def Quadratic(x, return_g=True, return_H=True):
        f = 0.5 * x.dot(A.dot(x)) + b.dot(x) + c
        out = (f,)
        if return_g:
            g = A.dot(x) + b
            out += (g,)
        if return_H:
            H = A
            out += (H,)
        return out if len(out) > 1 else out[0]

    return Quadratic


def assert_isadjoint(
    forward,
    adjoint,
    shape_u,
    shape_v,
    complex_u=False,
    complex_v=False,
    clinear=True,
    rtol=1e-6,
    atol=0.0,
    assert_error=True,
    random_seed=None,
):
    r"""Do a dot product test for the forward operator and its adjoint operator.

    Dot product test to verify the correctness of the adjoint operator
    :math:`F^H` of the forward operator :math:`F`.

    .. math::

        \mathbf{v}^H ( \mathbf{F} \mathbf{u} ) =
        ( \mathbf{F}^H \mathbf{v} )^H \mathbf{u}


    Parameters
    ----------
    forward : callable
        Forward operator.

    adjoint : callable
        Adjoint operator.

    shape_u : int, tuple of int
        Shape of vector ``u`` passed in to ``forward``; it is accordingly the
        expected shape of the vector returned from the ``adjoint``.

    shape_v : int, tuple of int
        Shape of vector ``v`` passed in to ``adjoint``; it is accordingly the
        expected shape of the vector returned from the ``forward``.

    complex_u : bool, default: False
        If True, vector ``u`` passed to ``forward`` is a complex vector;
        accordingly the ``adjoint`` is expected to return a complex vector.

    complex_v : bool, default: False
        If True, vector ``v`` passed to ``adjoint`` is a complex vector;
        accordingly the ``forward`` is expected to return a complex vector.

    clinear : bool, default: True
        If operator is complex-linear (True) or real-linear (False).

    rtol : float, default: 1e-6
        Relative tolerance.

    atol : float, default: 0.0
        Absolute tolerance.

    assert_error : bool, default: True
        By default this test is an assertion (silent if passed, raising an
        assertion error if failed). If set to False, the result of the test is
        returned as boolean and a message is printed.

    random_seed : numpy.random.Generator, int, optional
        The random number generator to use for the adjoint test. If an integer or None
        it is used to seed a new `numpy.random.default_rng`.

    Returns
    -------
    passed : bool, optional
        Result of the dot product test; only returned if ``assert_error`` is False.

    Raises
    ------
    AssertionError
        If the dot product test fails (only if assert_error=True).

    """
    __tracebackhide__ = True

    rng = np.random.default_rng(random_seed)

    def random(size, iscomplex):
        """Create random data of size and dtype of <size>."""
        out = rng.standard_normal(size)
        if iscomplex:
            out = out + 1j * rng.standard_normal(size)
        return out

    # Create random vectors u and v.
    u = random(np.prod(shape_u), complex_u).reshape(shape_u)
    v = random(np.prod(shape_v), complex_v).reshape(shape_v)

    # Carry out dot product test.
    fwd_u = forward(u)
    adj_v = adjoint(v)
    if clinear:
        lhs = np.vdot(v, fwd_u)  # lhs := v^H * (fwd * u)
        rhs = np.vdot(adj_v, u)  # rhs := (adj * v)^H * u
    else:
        lhs = np.vdot(v.real, fwd_u.real) + np.vdot(v.imag, fwd_u.imag)
        rhs = np.vdot(adj_v.real, u.real) + np.vdot(adj_v.imag, u.imag)

    # Check if they are the same.
    if assert_error:
        np.testing.assert_allclose(
            rhs, lhs, rtol=rtol, atol=atol, err_msg="Adjoint test failed"
        )

    else:
        passed = np.allclose(rhs, lhs, rtol=rtol, atol=atol)

        print(
            f"Adjoint test {'PASSED' if passed else 'FAILED'} ::  "
            f"{abs(rhs-lhs):.3e} < {atol+rtol*abs(lhs):.3e}  :: "
            f"|rhs-lhs| < atol + rtol|lhs|"
        )

        return passed


def assert_cell_intersects_geometric(
    cell, points, edges=None, faces=None, as_refine=False
):
    """Assert if a cell intersects a convex polygon.

    Parameters
    ----------
    cell : tree_mesh.TreeCell
        Must have cell.origin and cell.h properties
    points : (*, dim) array_like
        The points of the geometric object.
    edges : (*, 2) array_like of int, optional
        The 2 indices into points defining each edge
    faces : (*, 3) array_like of int, optional
        The 3 indices into points which lie on each face. These are used
        to define the face normals from the three points as
        ``norm = cross(p1 - p0, p2 - p0)``.
    as_refine : bool, or int
        If ``True`` (or a nonzero integer), this function will not assert and instead
        return either 0, -1, or the integer making it suitable (but slow) for
        refining a TreeMesh.

    Returns
    -------
    int

    Raises
    ------
    AssertionError
    """
    __tracebackhide__ = True

    x0 = cell.origin
    xF = x0 + cell.h

    points = np.atleast_2d(points)
    if edges is not None:
        edges = np.atleast_2d(edges)
        if edges.shape[-1] != 2:
            raise ValueError("Last dimension of edges must be 2.")
    if faces is not None:
        faces = np.atleast_2d(faces)
        if faces.shape[-1] != 3:
            raise ValueError("Last dimension of faces must be 3.")

    do_asserts = not as_refine
    level = -1
    if as_refine and not isinstance(as_refine, bool):
        level = int(as_refine)

    dim = points.shape[-1]
    # first the bounding box tests (associated with the 3 face normals of the cell
    mins = points.min(axis=0)
    for i_d in range(dim):
        if do_asserts:
            assert mins[i_d] <= xF[i_d]
        else:
            if mins[i_d] > xF[i_d]:
                return 0

    maxs = points.max(axis=0)
    for i_d in range(dim):
        if do_asserts:
            assert maxs[i_d] >= x0[i_d]
        else:
            if maxs[i_d] < x0[i_d]:
                return 0

    # create array of all the box points
    if edges is not None or faces is not None:
        box_points = np.meshgrid(*list(zip(x0, xF)))
        box_points = np.stack(box_points, axis=-1).reshape(-1, dim)

        def project_min_max(points, axis):
            ps = points @ axis
            return ps.min(), ps.max()

        if edges is not None and dim > 1:
            box_dirs = np.eye(dim)
            edge_dirs = points[edges[:, 1]] - points[edges[:, 0]]
            # perform the edge-edge intersection tests
            # these project all points onto the axis formed by the cross
            # product of the geometric edges and the bounding box's edges/faces normals
            for i in range(edges.shape[0]):
                for j in range(dim):
                    if dim == 3:
                        axis = np.cross(edge_dirs[i], box_dirs[j])
                    else:
                        axis = [-edge_dirs[i, 1], edge_dirs[i, 0]]
                    bmin, bmax = project_min_max(box_points, axis)
                    gmin, gmax = project_min_max(points, axis)
                    if do_asserts:
                        assert bmax >= gmin and bmin <= gmax
                    else:
                        if bmax < gmin or bmin > gmax:
                            return 0

        if faces is not None and dim > 2:
            face_normals = np.cross(
                points[faces[:, 1]] - points[faces[:, 0]],
                points[faces[:, 2]] - points[faces[:, 0]],
            )
            for i in range(faces.shape[0]):
                bmin, bmax = project_min_max(box_points, face_normals[i])
                gmin, gmax = project_min_max(points, face_normals[i])
                if do_asserts:
                    assert bmax >= gmin and bmin <= gmax
                else:
                    if bmax < gmin or bmin > gmax:
                        return 0
    if not do_asserts:
        return level


# DEPRECATIONS
setupMesh = deprecate_function(
    setup_mesh, "setupMesh", removal_version="1.0.0", error=True
)
Rosenbrock = deprecate_function(
    rosenbrock, "Rosenbrock", removal_version="1.0.0", error=True
)
checkDerivative = deprecate_function(
    check_derivative, "checkDerivative", removal_version="1.0.0", error=True
)
getQuadratic = deprecate_function(
    get_quadratic, "getQuadratic", removal_version="1.0.0", error=True
)



# distutils: language=c++
# cython: embedsignature=True, language_level=3
# cython: linetrace=True

from libcpp.pair cimport pair
from libcpp.unordered_map cimport unordered_map
from cython.operator cimport dereference
cimport cython
cimport numpy as np
from cython cimport view
from libc.math cimport sqrt

cdef extern from "triplet.h":
    cdef cppclass triplet[T, U, V]:
        T v1
        U v2
        V v3
        triplet()
        triplet(T, U, V)
import numpy as np

ctypedef fused ints:
    size_t
    np.int32_t
    np.int64_t

ctypedef fused pointers:
    size_t
    np.intp_t
    np.int32_t
    np.int64_t

@cython.boundscheck(False)
def _build_faces_edges(ints[:, :] simplices):
    # the node index in each simplex must be in increasing order
    cdef:
        int dim = simplices.shape[1] - 1
        ints n_simplex = simplices.shape[0]
        ints[:] simplex
        ints v1, v2, v3

        unordered_map[pair[ints, ints], ints] edges
        pair[ints, ints] edge
        ints n_edges = 0
        ints edges_per_simplex = 3 if dim==2 else 6
        ints[:, :] simplex_edges

        unordered_map[triplet[ints, ints, ints], ints] faces
        triplet[ints, ints, ints] face
        ints n_faces = 0
        ints faces_per_simplex = dim + 1
        ints[:, :] simplex_faces

    if ints is size_t:
        int_type = np.uintp
    elif ints is np.int32_t:
        int_type = np.int32
    elif ints is np.int64_t:
        int_type = np.int64

    simplex_edges = np.empty((n_simplex, edges_per_simplex), dtype=int_type)

    if dim == 3:
        simplex_faces = np.empty((n_simplex, faces_per_simplex), dtype=int_type)
    else:
        simplex_faces = simplex_edges

    cdef ints[:,:] edge_pairs = np.array(
        [[1, 2], [0, 2], [0, 1], [0, 3], [1, 3], [2, 3]],
        dtype=int_type
    )

    for i_simp in range(n_simplex):
        simplex = simplices[i_simp]

        # build edges
        for i_edge in range(edges_per_simplex):

            v1 = simplex[edge_pairs[i_edge, 0]]
            v2 = simplex[edge_pairs[i_edge, 1]]
            edge = pair[ints, ints](v1, v2)

            edge_search = edges.find(edge)
            if edge_search != edges.end():
                ind = dereference(edge_search).second
            else:
                ind = n_edges
                edges[edge] = ind
                n_edges += 1
            simplex_edges[i_simp, i_edge] = ind

        # build faces in 3D
        if dim == 3:
            for i_face in range(4):
                if i_face == 0:
                    v1 = simplex[1]
                    v2 = simplex[2]
                    v3 = simplex[3]
                elif i_face == 1:
                    v1 = simplex[0]
                    v2 = simplex[2]
                    v3 = simplex[3]
                elif i_face == 2:
                    v1 = simplex[0]
                    v2 = simplex[1]
                    v3 = simplex[3]
                else:
                    v1 = simplex[0]
                    v2 = simplex[1]
                    v3 = simplex[2]
                face = triplet[ints, ints, ints](v1, v2, v3)

                face_search = faces.find(face)
                if face_search != faces.end():
                    ind = dereference(face_search).second
                else:
                    ind = n_faces
                    faces[face] = ind
                    n_faces += 1
                simplex_faces[i_simp, i_face] = ind

    cdef ints[:, :] _edges = np.empty((n_edges, 2), dtype=int_type)
    for edge_it in edges:
        _edges[edge_it.second, 0] = edge_it.first.first
        _edges[edge_it.second, 1] = edge_it.first.second

    cdef ints[:, :] _faces
    if dim == 3:
        _faces =  np.empty((n_faces, 3), dtype=int_type)
        for face_it in faces:
            _faces[face_it.second, 0] = face_it.first.v1
            _faces[face_it.second, 1] = face_it.first.v2
            _faces[face_it.second, 2] = face_it.first.v3
    else:
        _faces = _edges

    cdef ints[:, :] face_edges
    cdef ints[:] _face

    if dim == 3:
        face_edges = np.empty((n_faces, 3), dtype=int_type)
        for i_face in range(n_faces):
            _face = _faces[i_face]
            # get indices of each edge in the face
            # 3 edges per face
            for i_edge in range(3):
                if i_edge == 0:
                    v1 = _face[1]
                    v2 = _face[2]
                elif i_edge == 1:
                    v1 = _face[0]
                    v2 = _face[2]
                elif i_edge == 2:
                    v1 = _face[0]
                    v2 = _face[1]
                # because of how faces were constructed, v1 < v2 always
                edge = pair[ints, ints](v1, v2)
                ind = edges[edge]
                face_edges[i_face, i_edge] = ind
    else:
        face_edges = np.empty((1, 1), dtype=int_type)

    return simplex_faces, _faces, simplex_edges, _edges, face_edges

@cython.boundscheck(False)
def _build_adjacency(ints[:, :] simplex_faces, n_faces):
    cdef:
        size_t n_cells = simplex_faces.shape[0]
        int dim = simplex_faces.shape[1] - 1
        np.int64_t[:, :] neighbors
        np.int64_t[:] visited
        ints[:] simplex

        ints i_cell, j, k, i_face, i_other

    if ints is size_t:
        int_type = np.uintp
    elif ints is np.int32_t:
        int_type = np.int32
    elif ints is np.int64_t:
        int_type = np.int64

    neighbors = np.full((n_cells, dim + 1), -1, dtype=np.int64)
    visited = np.full((n_faces), -1, dtype=np.int64)

    for i_cell in range(n_cells):
        simplex = simplex_faces[i_cell]
        for j in range(dim + 1):
            i_face = simplex[j]
            i_other = visited[i_face]
            if i_other == -1:
                visited[i_face] = i_cell
            else:
                neighbors[i_cell, j] = i_other
                k = 0
                while (k < dim + 1) and (simplex_faces[i_other, k] != i_face):
                    k += 1
                neighbors[i_other, k] = i_cell
    return neighbors

@cython.boundscheck(False)
@cython.linetrace(False)
cdef void _compute_bary_coords(
    np.float64_t[:] point,
    np.float64_t[:, :] Tinv,
    np.float64_t[:] shift,
    np.float64_t * bary
) nogil:
    cdef:
        int dim = point.shape[0]
        int i, j

    bary[dim] = 1.0
    for i in range(dim):
        bary[i] = 0.0
        for j in range(dim):
            bary[i] += Tinv[i, j] * (point[j] - shift[j])
        bary[dim] -= bary[i]

@cython.boundscheck(False)
def _directed_search(
    np.float64_t[:, :] locs,
    pointers[:] nearest_cc,
    np.float64_t[:, :] nodes,
    ints[:, :] simplex_nodes,
    np.int64_t[:, :] neighbors,
    np.float64_t[:, :, :] transform,
    np.float64_t[:, :] shift,
    np.float64_t eps=1E-15,
    bint zeros_outside=False,
    bint return_bary=True
):
    cdef:
        int i, j
        pointers i_simp
        int n_locs = locs.shape[0], dim = locs.shape[1]
        int max_directed = 1 + simplex_nodes.shape[0] // 4
        int i_directed
        bint is_inside
        np.int64_t[:] inds = np.full(len(locs), -1, dtype=np.int64)
        np.float64_t[:, :] all_barys = np.empty((1, 1), dtype=np.float64)
        np.float64_t barys[4]
        np.float64_t[:] loc
        np.float64_t[:, :] Tinv
        np.float64_t[:] rD
    if return_bary:
        all_barys = np.empty((len(locs), dim+1), dtype=np.float64)

    for i in range(n_locs):
        loc = locs[i]

        i_simp = nearest_cc[i]  # start at the nearest cell center
        i_directed = 0
        while i_directed < max_directed:
            Tinv = transform[i_simp]
            rD = shift[i_simp]
            _compute_bary_coords(loc, Tinv, rD, barys)
            j = 0
            is_inside = True
            while j <= dim:
                if barys[j] < -eps:
                    is_inside = False
                    # if not -1, move towards neighbor
                    if neighbors[i_simp, j] != -1:
                        i_simp = neighbors[i_simp, j]
                        break
                j += 1
            # If inside, I found my container
            if is_inside:
                break
            # Else, if I cycled through every bary
            # without breaking out of the above loop, that means I'm completely outside
            elif j == dim + 1:
                if zeros_outside:
                    i_simp = -1
                break
            i_directed += 1

        if i_directed == max_directed:
            # made it through the whole loop without breaking out
            # Mark as failed
            i_simp = -2
        inds[i] = i_simp
        if return_bary:
            for j in range(dim+1):
                all_barys[i, j] = barys[j]

    if return_bary:
        return np.array(inds), np.array(all_barys)
    return np.array(inds)

@cython.boundscheck(False)
@cython.cdivision(True)
def _interp_cc(
  np.float64_t[:, :] locs,
  np.float64_t[:, :] cell_centers,
  np.float64_t[:] mat_data,
  ints[:] mat_indices,
  ints[:] mat_indptr,
):
    cdef:
        ints i, j, diff, start, stop, i_d
        ints n_max_per_row = 0
        ints[:] close_cells
        int dim = locs.shape[1]
        np.float64_t[:, :] drs
        np.float64_t[:] rs
        np.float64_t[:] rhs
        np.float64_t[:] lambs
        np.float64_t[:] weights
        np.float64_t[:] point
        np.float64_t[:] close_cell
        np.float64_t det, weight_sum
        np.float64_t xx, xy, xz, yy, yz, zz
        bint too_close
        np.float64_t eps = 1E-15
    # Find maximum number per row to pre-allocate a storage
    for i in range(locs.shape[0]):
        diff = mat_indptr[i+1] - mat_indptr[i]
        if diff > n_max_per_row:
          n_max_per_row = diff
    #
    drs = np.empty((n_max_per_row, dim), dtype=np.float64)
    rs = np.empty((n_max_per_row,), dtype=np.float64)
    rhs = np.empty((dim,),dtype=np.float64)
    lambs = np.empty((dim,),dtype=np.float64)

    for i in range(locs.shape[0]):
        point = locs[i]
        start = mat_indptr[i]
        stop = mat_indptr[i+1]
        diff = stop-start
        close_cells = mat_indices[start:stop]
        for j in range(diff):
            rs[j] = 0.0
            close_cell = cell_centers[close_cells[j]]
            for i_d in range(dim):
                drs[j, i_d] = close_cell[i_d] - point[i_d]
                rs[j] += drs[j, i_d]*drs[j, i_d]
            rs[j] = sqrt(rs[j])
        weights = mat_data[start:stop]
        weights[:] = 0.0
        too_close = False
        i_d = 0
        for j in range(diff):
            if rs[j] < eps:
                too_close = True
                i_d = j
        if too_close:
            weights[i_d] = 1.0
        else:
            for j in range(diff):
                for i_d in range(dim):
                    drs[j, i_d] /= rs[j]
            xx = xy = yy = 0.0
            rhs[:] = 0.0
            if dim == 2:
                for j in range(diff):
                    xx += drs[j, 0] * drs[j, 0]
                    xy += drs[j, 0] * drs[j, 1]
                    yy += drs[j, 1] * drs[j, 1]
                    rhs[0] -= drs[j, 0]
                    rhs[1] -= drs[j, 1]
                det = xx * yy - xy * xy
                lambs[0] = (yy * rhs[0] - xy * rhs[1])/det
                lambs[1] = (-xy * rhs[0] + xx * rhs[1])/det

            if dim == 3:
                zz = xz = yz = 0.0
                for j in range(diff):
                    xx += drs[j, 0] * drs[j, 0]
                    xy += drs[j, 0] * drs[j, 1]
                    yy += drs[j, 1] * drs[j, 1]
                    xz += drs[j, 0] * drs[j, 2]
                    yz += drs[j, 1] * drs[j, 2]
                    zz += drs[j, 2] * drs[j, 2]
                    rhs[0] -= drs[j, 0]
                    rhs[1] -= drs[j, 1]
                    rhs[2] -= drs[j, 2]

                det = (
                      xx * (yy * zz - yz * yz)
                      + xy * (xz * yz - xy * zz)
                      + xz * (xy * yz - xz * yy)
                )
                lambs[0] = (
                    (yy * zz - yz * yz) * rhs[0]
                    + (xz * yz - xy * zz) * rhs[1]
                    + (xy * yz - xz * yy) * rhs[2]
                )/det
                lambs[1] = (
                    (xz * yz - xy * zz) * rhs[0]
                    + (xx * zz - xz * xz) * rhs[1]
                    + (xy * xz - xx * yz) * rhs[2]
                )/det
                lambs[2] = (
                    (xy * yz - xz * yy) * rhs[0]
                    + (xz * xy - xx * yz) * rhs[1]
                    + (xx * yy - xy * xy) * rhs[2]
                )/det

            weight_sum = 0.0
            for j in range(diff):
                weights[j] = 1.0
                for i_d in range(dim):
                    weights[j] += lambs[i_d] * drs[j, i_d]
                weights[j] /= rs[j]
                weight_sum += weights[j]
            for j in range(diff):
                weights[j] /= weight_sum



"""Module containing the TreeMesh implementation."""

#      ___          ___       ___          ___          ___          ___
#     /\  \        /\  \     /\  \        /\  \        /\  \        /\  \
#    /::\  \      /::\  \    \:\  \      /::\  \      /::\  \      /::\  \
#   /:/\:\  \    /:/\:\  \    \:\  \    /:/\:\  \    /:/\:\  \    /:/\:\  \
#  /:/  \:\  \  /:/  \:\  \   /::\  \  /::\~\:\  \  /::\~\:\  \  /::\~\:\  \
# /:/__/ \:\__\/:/__/ \:\__\ /:/\:\__\/:/\:\ \:\__\/:/\:\ \:\__\/:/\:\ \:\__\
# \:\  \ /:/  /\:\  \  \/__//:/  \/__/\/_|::\/:/  /\:\~\:\ \/__/\:\~\:\ \/__/
#  \:\  /:/  /  \:\  \     /:/  /        |:|::/  /  \:\ \:\__\   \:\ \:\__\
#   \:\/:/  /    \:\  \    \/__/         |:|\/__/    \:\ \/__/    \:\ \/__/
#    \::/  /      \:\__\                 |:|  |       \:\__\       \:\__\
#     \/__/        \/__/                  \|__|        \/__/        \/__/
#
#
#
#                      .----------------.----------------.
#                     /|               /|               /|
#                    / |              / |              / |
#                   /  |      6      /  |     7       /  |
#                  /   |            /   |            /   |
#                 .----------------.----+-----------.    |
#                /|    . ---------/|----.----------/|----.
#               / |   /|         / |   /|         / |   /|
#              /  |  / |  4     /  |  / |   5    /  |  / |
#             /   | /  |       /   | /  |       /   | /  |
#            . -------------- .----------------.    |/   |
#            |    . ---+------|----.----+------|----.    |
#            |   /|    .______|___/|____.______|___/|____.
#            |  / |   /    2  |  / |   /     3 |  / |   /
#            | /  |  /        | /  |  /        | /  |  /
#            . ---+---------- . ---+---------- .    | /
#            |    |/          |    |/          |    |/             z
#            |    . ----------|----.-----------|----.              ^   y
#            |   /      0     |   /       1    |   /               |  /
#            |  /             |  /             |  /                | /
#            | /              | /              | /                 o----> x
#            . -------------- . -------------- .
#
#
# Face Refinement:
#
#      2_______________3                    _______________
#      |               |                   |       |       |
#   ^  |               |                   |   2   |   3   |
#   |  |               |                   |       |       |
#   |  |       x       |        --->       |-------+-------|
#   t1 |               |                   |       |       |
#      |               |                   |   0   |   1   |
#      |_______________|                   |_______|_______|
#      0      t0-->    1
#
#
# Face and Edge naming conventions:
#
#                      fZp
#                       |
#                 6 ------eX3------ 7
#                /|     |         / |
#               /eZ2    .        / eZ3
#             eY2 |        fYp eY3  |
#             /   |            / fXp|
#            4 ------eX2----- 5     |
#            |fXm 2 -----eX1--|---- 3          z
#           eZ0  /            |  eY1           ^   y
#            | eY0   .  fYm  eZ1 /             |  /
#            | /     |        | /              | /
#            0 ------eX0------1                o----> x
#                    |
#                   fZm
#
#
#            fX                                  fY
#      2___________3                       2___________3
#      |     e1    |                       |     e1    |
#      |           |                       |           |
#   e0 |     x     | e2      z          e0 |     x     | e2      z
#      |           |         ^             |           |         ^
#      |___________|         |___> y       |___________|         |___> x
#      0    e3     1                       0    e3     1
#           fZ
#      2___________3
#      |     e1    |
#      |           |
#   e0 |     x     | e2      y
#      |           |         ^
#      |___________|         |___> x
#      0    e3     1

from discretize.base import BaseTensorMesh
from discretize.operators import InnerProducts, DiffOperators
from discretize.mixins import InterfaceMixins, TreeMeshIO
from discretize._extensions.tree_ext import _TreeMesh, TreeCell  # NOQA F401
import numpy as np
import scipy.sparse as sp
from discretize.utils.code_utils import deprecate_property
from scipy.spatial import Delaunay


class TreeMesh(
    _TreeMesh,
    InnerProducts,
    DiffOperators,
    BaseTensorMesh,
    TreeMeshIO,
    InterfaceMixins,
):
    """Class for QuadTree (2D) and OcTree (3D) meshes.

    Tree meshes are numerical grids where the dimensions of each cell are powers of 2
    larger than some base cell dimension. Unlike the :class:`~discretize.TensorMesh`
    class, gridded locations and numerical operators for instances of ``TreeMesh``
    cannot be simply constructed using tensor products. Furthermore, each cell
    is an instance of :class:`~discretize.tree_mesh.TreeCell` .

    Each cell of a `TreeMesh` has a certain level associated with it which describes its
    height in the structure of the tree. The tree mesh is refined by specifying what
    level you want in a given location. They start at level 0, defining the largest
    possible cell on the mesh, and go to a level of `max_level` describing the
    finest possible cell on the mesh. TreeMesh` contains several refinement functions
    used to design the mesh, starting with a general purpose refinement function, along
    with several faster specialized refinement functions based on fundamental geometric
    entities:

    - `refine` -> The general purpose refinement function supporting arbitrary defined functions
    - `insert_cells`
    - `refine_ball`
    - `refine_line`
    - `refine_triangle`
    - `refine_box`
    - `refine_tetrahedron`
    - `refine_vertical_trianglular_prism`
    - `refine_bounding_box`
    - `refine_points`
    - `refine_surface`

    Like array indexing in python, you can also supply negative indices as a level
    arguments to these functions to index levels in a reveresed order (i.e. -1 is
    equivalent to `max_level`).

    Parameters
    ----------
    h : (dim) iterable of int, numpy.ndarray, or tuple
        Defines the cell widths of the *underlying tensor mesh* along each axis. The
        length of the iterable object is equal to the dimension of the mesh (2 or 3).
        For a 3D mesh, the list would have the form *[hx, hy, hz]*. The number of cells
        along each axis **must be a power of 2** .

        Along each axis, the user has 3 choices for defining the cells widths for the
        underlying tensor mesh:

        - :class:`int` -> A unit interval is equally discretized into `N` cells.
        - :class:`numpy.ndarray` -> The widths are explicity given for each cell
        - the widths are defined as a :class:`list` of :class:`tuple` of the form *(dh, nc, [npad])*
          where *dh* is the cell width, *nc* is the number of cells, and *npad* (optional)
          is a padding factor denoting exponential increase/decrease in the cell width
          for each cell; e.g. *[(2., 10, -1.3), (2., 50), (2., 10, 1.3)]*

    origin : (dim) iterable, default: 0
        Define the origin or 'anchor point' of the mesh; i.e. the bottom-left-frontmost
        corner. By default, the mesh is anchored such that its origin is at [0, 0, 0].

        For each dimension (x, y or z), The user may set the origin 2 ways:

        - a ``scalar`` which explicitly defines origin along that dimension.
        - **{'0', 'C', 'N'}** a :class:`str` specifying whether the zero coordinate along
          each axis is the first node location ('0'), in the center ('C') or the last
          node location ('N') (see Examples).

    diagonal_balance : bool, optional
        Whether to balance cells along the diagonal of the tree during construction.
        This will effect all calls to refine the tree.

    Examples
    --------
    Here we generate a basic 2D tree mesh.

    >>> from discretize import TreeMesh
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt

    Define base mesh (domain and finest discretization),

    >>> dh = 5    # minimum cell width (base mesh cell width)
    >>> nbc = 64  # number of base mesh cells
    >>> h = dh * np.ones(nbc)
    >>> mesh = TreeMesh([h, h])

    Define corner points for a rectangular box, and subdived the mesh within the box
    to the maximum refinement level.

    >>> x0s = [120.0, 80.0]
    >>> x1s = [240.0, 160.0]
    >>> levels = [mesh.max_level]
    >>> mesh.refine_box(x0s, x1s, levels)

    >>> mesh.plot_grid()
    >>> plt.show()
    """

    _meshType = "TREE"
    _aliases = {
        **BaseTensorMesh._aliases,
        **DiffOperators._aliases,
        **{
            "ntN": "n_total_nodes",
            "ntEx": "n_total_edges_x",
            "ntEy": "n_total_edges_y",
            "ntEz": "n_total_edges_z",
            "ntE": "n_total_edges",
            "ntFx": "n_total_faces_x",
            "ntFy": "n_total_faces_y",
            "ntFz": "n_total_faces_z",
            "ntF": "n_total_faces",
            "nhN": "n_hanging_nodes",
            "nhEx": "n_hanging_edges_x",
            "nhEy": "n_hanging_edges_y",
            "nhEz": "n_hanging_edges_z",
            "nhE": "n_hanging_edges",
            "nhFx": "n_hanging_faces_x",
            "nhFy": "n_hanging_faces_y",
            "nhFz": "n_hanging_faces_z",
            "nhF": "n_hanging_faces",
            "gridhN": "hanging_nodes",
            "gridhFx": "hanging_faces_x",
            "gridhFy": "hanging_faces_y",
            "gridhFz": "hanging_faces_z",
            "gridhEx": "hanging_edges_x",
            "gridhEy": "hanging_edges_y",
            "gridhEz": "hanging_edges_z",
        },
    }
    _items = {"h", "origin", "cell_state"}

    # inheriting stuff from BaseTensorMesh that isn't defined in _QuadTree
    def __init__(self, h=None, origin=None, diagonal_balance=False, **kwargs):
        if "x0" in kwargs:
            origin = kwargs.pop("x0")
        super().__init__(h=h, origin=origin, diagonal_balance=diagonal_balance)

        cell_state = kwargs.pop("cell_state", None)
        cell_indexes = kwargs.pop("cell_indexes", None)
        cell_levels = kwargs.pop("cell_levels", None)
        if cell_state is None:
            if cell_indexes is not None and cell_levels is not None:
                cell_state = {}
                cell_state["indexes"] = cell_indexes
                cell_state["levels"] = cell_levels
        if cell_state is not None:
            indexes = cell_state["indexes"]
            levels = cell_state["levels"]
            self.__setstate__((indexes, levels))

    def __repr__(self):
        """Plain text representation."""
        mesh_name = "{0!s}TreeMesh".format(("Oc" if self.dim == 3 else "Quad"))

        top = "\n" + mesh_name + ": {0:2.2f}% filled\n\n".format(self.fill * 100)

        # Number of cells per level
        level_count = self._count_cells_per_index()
        non_zero_levels = np.nonzero(level_count)[0]
        cell_display = ["Level : Number of cells"]
        cell_display.append("-----------------------")
        for level in non_zero_levels:
            cell_display.append("{:^5} : {:^15}".format(level, level_count[level]))
        cell_display.append("-----------------------")
        cell_display.append("Total : {:^15}".format(self.nC))

        extent_display = ["            Mesh Extent       "]
        extent_display.append("        min     ,     max     ")
        extent_display.append("   ---------------------------")
        dim_label = {0: "x", 1: "y", 2: "z"}
        for dim in range(self.dim):
            n_vector = getattr(self, "nodes_" + dim_label[dim])
            extent_display.append(
                "{}: {:^13},{:^13}".format(dim_label[dim], n_vector[0], n_vector[-1])
            )

        for i, line in enumerate(extent_display):
            if i == len(cell_display):
                cell_display.append(" " * (len(cell_display[0]) - 3 - len(line)))
            cell_display[i] += 3 * " " + line

        h_display = ["     Cell Widths    "]
        h_display.append("    min   ,   max   ")
        h_display.append("-" * (len(h_display[0])))
        h_gridded = self.h_gridded
        mins = np.min(h_gridded, axis=0)
        maxs = np.max(h_gridded, axis=0)
        for dim in range(self.dim):
            h_display.append("{:^10}, {:^10}".format(mins[dim], maxs[dim]))

        for i, line in enumerate(h_display):
            if i == len(cell_display):
                cell_display.append(" " * len(cell_display[0]))
            cell_display[i] += 3 * " " + line

        return top + "\n".join(cell_display)

    def _repr_html_(self):
        """HTML representation."""
        mesh_name = "{0!s}TreeMesh".format(("Oc" if self.dim == 3 else "Quad"))
        level_count = self._count_cells_per_index()
        non_zero_levels = np.nonzero(level_count)[0]
        dim_label = {0: "x", 1: "y", 2: "z"}
        h_gridded = self.h_gridded
        mins = np.min(h_gridded, axis=0)
        maxs = np.max(h_gridded, axis=0)

        style = " style='padding: 5px 20px 5px 20px;'"
        # Cell level table:
        cel_tbl = "<table>\n"
        cel_tbl += "<tr>\n"
        cel_tbl += "<th" + style + ">Level</th>\n"
        cel_tbl += "<th" + style + ">Number of cells</th>\n"
        cel_tbl += "</tr>\n"
        for level in non_zero_levels:
            cel_tbl += "<tr>\n"
            cel_tbl += "<td" + style + ">{}</td>\n".format(level)
            cel_tbl += "<td" + style + ">{}</td>\n".format(level_count[level])
            cel_tbl += "</tr>\n"
        cel_tbl += "<tr>\n"
        cel_tbl += (
            "<td style='font-weight: bold; padding: 5px 20px 5px 20px;'> Total </td>\n"
        )
        cel_tbl += "<td" + style + "> {} </td>\n".format(self.nC)
        cel_tbl += "</tr>\n"
        cel_tbl += "</table>\n"

        det_tbl = "<table>\n"
        det_tbl += "<tr>\n"
        det_tbl += "<th></th>\n"
        det_tbl += "<th" + style + " colspan='2'>Mesh extent</th>\n"
        det_tbl += "<th" + style + " colspan='2'>Cell widths</th>\n"
        det_tbl += "</tr>\n"

        det_tbl += "<tr>\n"
        det_tbl += "<th></th>\n"
        det_tbl += "<th" + style + ">min</th>\n"
        det_tbl += "<th" + style + ">max</th>\n"
        det_tbl += "<th" + style + ">min</th>\n"
        det_tbl += "<th" + style + ">max</th>\n"
        det_tbl += "</tr>\n"
        for dim in range(self.dim):
            n_vector = getattr(self, "nodes_" + dim_label[dim])
            det_tbl += "<tr>\n"
            det_tbl += "<td" + style + ">{}</td>\n".format(dim_label[dim])
            det_tbl += "<td" + style + ">{}</td>\n".format(n_vector[0])
            det_tbl += "<td" + style + ">{}</td>\n".format(n_vector[-1])
            det_tbl += "<td" + style + ">{}</td>\n".format(mins[dim])
            det_tbl += "<td" + style + ">{}</td>\n".format(maxs[dim])
            det_tbl += "</tr>\n"
        det_tbl += "</table>\n"

        full_tbl = "<table>\n"
        full_tbl += "<tr>\n"
        full_tbl += "<td style='font-weight: bold; font-size: 1.2em; text-align: center;'>{}</td>\n".format(
            mesh_name
        )
        full_tbl += "<td style='font-size: 1.2em; text-align: center;' colspan='2'>{0:2.2f}% filled</td>\n".format(
            100 * self.fill
        )
        full_tbl += "</tr>\n"
        full_tbl += "<tr>\n"

        full_tbl += "<td>\n"
        full_tbl += cel_tbl
        full_tbl += "</td>\n"

        full_tbl += "<td>\n"
        full_tbl += det_tbl
        full_tbl += "</td>\n"

        full_tbl += "</tr>\n"
        full_tbl += "</table>\n"

        return full_tbl

    @BaseTensorMesh.origin.setter
    def origin(self, value):  # NOQA D102
        # first use the BaseTensorMesh to set the origin to handle "0, C, N"
        BaseTensorMesh.origin.fset(self, value)
        # then update the TreeMesh with the hidden value
        self._set_origin(self._origin)

    def refine_bounding_box(
        self,
        points,
        level=-1,
        padding_cells_by_level=None,
        finalize=True,
        diagonal_balance=None,
    ):
        """Refine within a bounding box based on the maximum and minimum extent of scattered points.

        This function refines the tree mesh based on the bounding box defined by the
        maximum and minimum extent of the scattered input `points`. It will refine
        the tree mesh for each level given.

        It also optionally pads the bounding box at each level based on the number of
        padding cells at each dimension.

        Parameters
        ----------
        points : (N, dim) array_like
            The bounding box will be the maximum and minimum extent of these points
        level : int, optional
            The level of the treemesh to refine the bounding box to.
            Negative values index tree levels backwards, (e.g. `-1` is `max_level`).
        padding_cells_by_level : None, int, (n_level) array_like or (n_level, dim) array_like, optional
            The number of cells to pad the bounding box at each level of refinement.
            If a single number, each level below ``level`` will be padded with those
            number of cells. If array_like, `n_level`s below ``level`` will be padded,
            where if a 1D array, each dimension will be padded with the same number of
            cells, or 2D array supports variable padding along each dimension. None
            implies no padding.
        finalize : bool, optional
            Whether to finalize the mesh after the call.
        diagonal_balance : None or bool, optional
            Whether to balance cells diagonally in the refinement, `None` implies using
            the same setting used to instantiate the TreeMesh`.

        See Also
        --------
        refine_box

        Examples
        --------
        Given a set of points, we want to refine the tree mesh with the bounding box
        that surrounds those points. The arbitrary points we use for this example are
        uniformly scattered between [3/8, 5/8] in the first and second dimension.

        >>> import discretize
        >>> import matplotlib.pyplot as plt
        >>> import matplotlib.patches as patches
        >>> mesh = discretize.TreeMesh([32, 32])
        >>> rng = np.random.default_rng(852)
        >>> points = rng.random((20, 2)) * 0.25 + 3/8

        Now we want to refine to the maximum level, with no padding the in `x`
        direction and `2` cells in `y`. At the second highest level we want 2 padding
        cells in each direction beyond that.

        >>> padding = [[0, 2], [2, 2]]
        >>> mesh.refine_bounding_box(points, -1, padding)

        For demonstration we, overlay the bounding box to show the effects of padding.

        >>> ax = mesh.plot_grid()
        >>> rect = patches.Rectangle([3/8, 3/8], 1/4, 1/4, facecolor='none', edgecolor='r', linewidth=3)
        >>> ax.add_patch(rect)
        >>> plt.show()
        """
        bsw = np.min(np.atleast_2d(points), axis=0)
        tnw = np.max(np.atleast_2d(points), axis=0)
        if level < 0:
            level = (self.max_level + 1) - (abs(level) % (self.max_level + 1))
        if level > self.max_level:
            raise IndexError(f"Level beyond max octree level, {self.max_level}")

        # pad based on the number of cells at each level
        if padding_cells_by_level is None:
            padding_cells_by_level = np.array([0])
        padding_cells_by_level = np.asarray(padding_cells_by_level)
        if padding_cells_by_level.ndim == 0 and level > 1:
            padding_cells_by_level = np.full(level - 1, padding_cells_by_level)
        padding_cells_by_level = np.atleast_1d(padding_cells_by_level)
        if (
            padding_cells_by_level.ndim == 2
            and padding_cells_by_level.shape[1] != self.dim
        ):
            raise ValueError("incorrect dimension for padding_cells_by_level.")

        h_min = np.r_[[h.min() for h in self.h]]
        x0 = []
        xF = []
        ls = []
        # The zip will short circuit on the shorter array.
        for lv, n_pad in zip(np.arange(level, 1, -1), padding_cells_by_level):
            padding_at_level = n_pad * h_min * 2 ** (self.max_level - lv)
            bsw = bsw - padding_at_level
            tnw = tnw + padding_at_level
            x0.append(bsw)
            xF.append(tnw)
            ls.append(lv)

        self.refine_box(
            x0, xF, ls, finalize=finalize, diagonal_balance=diagonal_balance
        )

    def refine_points(
        self,
        points,
        level=-1,
        padding_cells_by_level=None,
        finalize=True,
        diagonal_balance=None,
    ):
        """Refine the mesh at given points to the prescribed level.

        This function refines the tree mesh around the `points`. It will refine
        the tree mesh for each level given. It also optionally radially pads around each
        point at each level with the number of padding cells given.

        Parameters
        ----------
        points : (N, dim) array_like
            The points to be refined around.
        level : int, optional
            The level of the tree mesh to refine to. Negative values index tree levels
            backwards, (e.g. `-1` is `max_level`).
        padding_cells_by_level : None, int, (n_level) array_like, optional
            The number of cells to pad the bounding box at each level of refinement.
            If a single number, each level below ``level`` will be padded with those
            number of cells. If array_like, `n_level`s below ``level`` will be padded.
            None implies no padding.
        finalize : bool, optional
            Whether to finalize the mesh after the call.
        diagonal_balance : None or bool, optional
            Whether to balance cells diagonally in the refinement, `None` implies using
            the same setting used to instantiate the TreeMesh`.

        See Also
        --------
        refine_ball, insert_cells

        Examples
        --------
        Given a set of points, we want to refine the tree mesh around these points to
        a prescribed level with a certain amount of padding.

        >>> import discretize
        >>> import matplotlib.pyplot as plt
        >>> import matplotlib.patches as patches
        >>> mesh = discretize.TreeMesh([32, 32])

        Now we want to refine to the maximum level with 1 cell padding around the point,
        and then refine at the second highest level with at least 3 cells beyond that.

        >>> points = np.array([[0.1, 0.3], [0.6, 0.8]])
        >>> padding = [1, 3]
        >>> mesh.refine_points(points, -1, padding)

        >>> ax = mesh.plot_grid()
        >>> ax.scatter(*points.T, color='C1')
        >>> plt.show()
        """
        if level < 0:
            level = (self.max_level + 1) - (abs(level) % (self.max_level + 1))
        if level > self.max_level:
            raise IndexError(f"Level beyond max octree level, {self.max_level}")

        # pad based on the number of cells at each level
        if padding_cells_by_level is None:
            padding_cells_by_level = np.array([0])
        padding_cells_by_level = np.asarray(padding_cells_by_level)
        if padding_cells_by_level.ndim == 0 and level > 1:
            padding_cells_by_level = np.full(level - 1, padding_cells_by_level)
        padding_cells_by_level = np.atleast_1d(padding_cells_by_level)

        h_min = np.max([h.min() for h in self.h])
        radius_at_level = 0.0
        for lv, n_pad in zip(np.arange(level, 1, -1), padding_cells_by_level):
            radius_at_level += n_pad * h_min * 2 ** (self.max_level - lv)
            if radius_at_level == 0:
                self.insert_cells(
                    points, lv, finalize=False, diagonal_balance=diagonal_balance
                )
            else:
                self.refine_ball(
                    points,
                    radius_at_level,
                    lv,
                    finalize=False,
                    diagonal_balance=diagonal_balance,
                )

        if finalize:
            self.finalize()

    def refine_surface(
        self,
        xyz,
        level=-1,
        padding_cells_by_level=None,
        pad_up=False,
        pad_down=True,
        finalize=True,
        diagonal_balance=None,
    ):
        """Refine along a surface triangulated from xyz to the prescribed level.

        This function refines the mesh based on a triangulated surface from the input
        points. Every cell that intersects the surface will be refined to the given
        level.

        It also optionally pads the surface at each level based on the number of
        padding cells at each dimension. It does so by stretching the bounding box of
        the surface in each dimension.

        Parameters
        ----------
        xyz : (N, dim) array_like or tuple
            The points defining the surface. Will be triangulated along the horizontal
            dimensions. You are able to supply your own triangulation in 3D by passing
            a tuple of (`xyz`, `triangle_indices`).
        level : int, optional
            The level of the tree mesh to refine to. Negative values index tree levels
            backwards, (e.g. `-1` is `max_level`).
        padding_cells_by_level : None, int, (n_level) array_like or (n_level, dim) array_like, optional
            The number of cells to pad the bounding box at each level of refinement.
            If a single number, each level below ``level`` will be padded with those
            number of cells. If array_like, `n_level`s below ``level`` will be padded,
            where if a 1D array, each dimension will be padded with the same number of
            cells, or 2D array supports variable padding along each dimension. None
            implies no padding.
        pad_up : bool, optional
            Whether to pad above the surface.
        pad_down : bool, optional
            Whether to pad below the surface.
        finalize : bool, optional
            Whether to finalize the mesh after the call.
        diagonal_balance : None or bool, optional
            Whether to balance cells diagonally in the refinement, `None` implies using
            the same setting used to instantiate the TreeMesh`.

        See Also
        --------
        refine_triangle, refine_vertical_trianglular_prism

        Examples
        --------
        In 2D we define the surface as a line segment, which we would like to refine
        along.

        >>> import discretize
        >>> import matplotlib.pyplot as plt
        >>> import matplotlib.patches as patches
        >>> mesh = discretize.TreeMesh([32, 32])

        This surface is a simple sine curve, which we would like to pad with at least
        2 cells vertically below at the maximum level, and 3 cells below that at the
        second highest level.

        >>> x = np.linspace(0.2, 0.8, 51)
        >>> z = 0.25*np.sin(2*np.pi*x)+0.5
        >>> xz = np.c_[x, z]
        >>> mesh.refine_surface(xz, -1, [[0, 2], [0, 3]])

        >>> ax = mesh.plot_grid()
        >>> ax.plot(x, z, color='C1')
        >>> plt.show()

        In 3D we define a grid of surface locations with there corresponding elevations.
        In this example we pad 2 cells at the finest level below the surface, and 3
        cells down at the next level.

        >>> mesh = discretize.TreeMesh([32, 32, 32])
        >>> x, y = np.mgrid[0.2:0.8:21j, 0.2:0.8:21j]
        >>> z = 0.125*np.sin(2*np.pi*x) + 0.5 + 0.125 * np.cos(2 * np.pi * y)
        >>> points = np.stack([x, y, z], axis=-1).reshape(-1, 3)
        >>> mesh.refine_surface(points, -1, [[0, 0, 2], [0, 0, 3]])

        >>> v = mesh.cell_levels_by_index(np.arange(mesh.n_cells))
        >>> fig, axs = plt.subplots(1, 3, figsize=(12,4))
        >>> mesh.plot_slice(v, ax=axs[0], normal='x', grid=True, clim=[2, 5])
        >>> mesh.plot_slice(v, ax=axs[1], normal='y', grid=True, clim=[2, 5])
        >>> mesh.plot_slice(v, ax=axs[2], normal='z', grid=True, clim=[2, 5])
        >>> plt.show()
        """
        if level < 0:
            level = (self.max_level + 1) - (abs(level) % (self.max_level + 1))
        if level > self.max_level:
            raise IndexError(f"Level beyond max octree level, {self.max_level}")

        # pad based on the number of cells at each level
        if padding_cells_by_level is None:
            padding_cells_by_level = np.array([0])
        padding_cells_by_level = np.asarray(padding_cells_by_level)
        if padding_cells_by_level.ndim == 0 and level > 1:
            padding_cells_by_level = np.full(level - 1, padding_cells_by_level)
        padding_cells_by_level = np.atleast_1d(padding_cells_by_level)
        if (
            padding_cells_by_level.ndim == 2
            and padding_cells_by_level.shape[1] != self.dim
        ):
            raise ValueError("incorrect dimension for padding_cells_by_level.")

        if self.dim == 2:
            xyz = np.asarray(xyz)
            sorter = np.argsort(xyz[:, 0])
            xyz = xyz[sorter]
            n_ps = len(xyz)
            inds = np.arange(n_ps)
            simps1 = np.c_[inds[:-1], inds[1:], inds[:-1]] + [0, 0, n_ps]
            simps2 = np.c_[inds[:-1], inds[1:], inds[1:]] + [n_ps, n_ps, 0]
            simps = np.r_[simps1, simps2]
        else:
            if isinstance(xyz, tuple):
                xyz, simps = xyz
                xyz = np.asarray(xyz)
                simps = np.asarray(simps)
            else:
                xyz = np.asarray(xyz)
                triang = Delaunay(xyz[:, :2])
                simps = triang.simplices
            n_ps = len(xyz)

        # calculate bounding box for padding
        bb_min = np.min(xyz, axis=0)[:-1]
        bb_max = np.max(xyz, axis=0)[:-1]
        half_width = (bb_max - bb_min) / 2
        center = (bb_max + bb_min) / 2
        points = np.empty((n_ps, self.dim))
        points[:, -1] = xyz[:, -1]

        h_min = np.r_[[h.min() for h in self.h]]
        pad = 0.0
        for lv, n_pad in zip(np.arange(level, 1, -1), padding_cells_by_level):
            pad += n_pad * h_min * 2 ** (self.max_level - lv)
            h = 0
            if pad_up:
                h += pad[-1]
            if pad_down:
                h += pad[-1]
                points[:, -1] = xyz[:, -1] - pad[-1]
            horizontal_expansion = (half_width + pad[:-1]) / half_width
            points[:, :-1] = horizontal_expansion * (xyz[:, :-1] - center) + center
            if self.dim == 2:
                triangles = np.r_[points, points + [0, h]][simps]
                self.refine_triangle(
                    triangles, lv, finalize=False, diagonal_balance=diagonal_balance
                )
            else:
                triangles = points[simps]
                self.refine_vertical_trianglular_prism(
                    triangles, h, lv, finalize=False, diagonal_balance=diagonal_balance
                )

        if finalize:
            self.finalize()

    @property
    def total_nodes(self):
        """Gridded hanging and non-hanging nodes locations.

        This property returns a numpy array of shape
        ``(n_total_nodes, dim)`` containing gridded locations for
        all hanging and non-hanging nodes in the mesh.

        Returns
        -------
        (n_total_nodes, dim) numpy.ndarray of float
            Gridded hanging and non-hanging node locations
        """
        return np.vstack((self.nodes, self.hanging_nodes))

    @property
    def vntF(self):
        """Vector number of total faces along each axis.

        This property returns the total number of hanging and
        non-hanging faces along each axis direction. The returned
        quantity is a list of integers of the form [nFx,nFy,nFz].

        Returns
        -------
        list of int
            Vector number of total faces along each axis
        """
        return [self.ntFx, self.ntFy] + ([] if self.dim == 2 else [self.ntFz])

    @property
    def vntE(self):
        """Vector number of total edges along each axis.

        This property returns the total number of hanging and
        non-hanging edges along each axis direction. The returned
        quantity is a list of integers of the form [nEx,nEy,nEz].

        Returns
        -------
        list of int
            Vector number of total edges along each axis
        """
        return [self.ntEx, self.ntEy] + ([] if self.dim == 2 else [self.ntEz])

    @property
    def stencil_cell_gradient(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if getattr(self, "_stencil_cell_gradient", None) is None:
            self._stencil_cell_gradient = sp.vstack(
                [self.stencil_cell_gradient_x, self.stencil_cell_gradient_y]
            )
            if self.dim == 3:
                self._stencil_cell_gradient = sp.vstack(
                    [self._stencil_cell_gradient, self.stencil_cell_gradient_z]
                )

        return self._stencil_cell_gradient

    @property
    def cell_gradient(self):  # NOQA D102
        # Documentation inherited from discretize.operators.DifferentialOperators
        if getattr(self, "_cell_gradient", None) is None:
            i_s = self.face_boundary_indices

            ix = np.ones(self.nFx)
            ix[i_s[0]] = 0.0
            ix[i_s[1]] = 0.0
            Pafx = sp.diags(ix)

            iy = np.ones(self.nFy)
            iy[i_s[2]] = 0.0
            iy[i_s[3]] = 0.0
            Pafy = sp.diags(iy)

            MfI = self.get_face_inner_product(invert_matrix=True)

            if self.dim == 2:
                Pi = sp.block_diag([Pafx, Pafy])

            elif self.dim == 3:
                iz = np.ones(self.nFz)
                iz[i_s[4]] = 0.0
                iz[i_s[5]] = 0.0
                Pafz = sp.diags(iz)
                Pi = sp.block_diag([Pafx, Pafy, Pafz])

            self._cell_gradient = (
                -Pi * MfI * self.face_divergence.T * sp.diags(self.cell_volumes)
            )

        return self._cell_gradient

    @property
    def cell_gradient_x(self):  # NOQA D102
        # Documentation inherited from discretize.operators.DifferentialOperators
        if getattr(self, "_cell_gradient_x", None) is None:
            nFx = self.nFx
            i_s = self.face_boundary_indices

            ix = np.ones(self.nFx)
            ix[i_s[0]] = 0.0
            ix[i_s[1]] = 0.0
            Pafx = sp.diags(ix)

            MfI = self.get_face_inner_product(invert_matrix=True)
            MfIx = sp.diags(MfI.diagonal()[:nFx])

            self._cell_gradient_x = (
                -Pafx * MfIx * self.face_x_divergence.T * sp.diags(self.cell_volumes)
            )

        return self._cell_gradient_x

    @property
    def cell_gradient_y(self):  # NOQA D102
        # Documentation inherited from discretize.operators.DifferentialOperators
        if getattr(self, "_cell_gradient_y", None) is None:
            nFx = self.nFx
            nFy = self.nFy
            i_s = self.face_boundary_indices

            iy = np.ones(self.nFy)
            iy[i_s[2]] = 0.0
            iy[i_s[3]] = 0.0
            Pafy = sp.diags(iy)

            MfI = self.get_face_inner_product(invert_matrix=True)
            MfIy = sp.diags(MfI.diagonal()[nFx : nFx + nFy])

            self._cell_gradient_y = (
                -Pafy * MfIy * self.face_y_divergence.T * sp.diags(self.cell_volumes)
            )

        return self._cell_gradient_y

    @property
    def cell_gradient_z(self):  # NOQA D102
        # Documentation inherited from discretize.operators.DifferentialOperators
        if self.dim == 2:
            raise TypeError("z derivative not defined in 2D")
        if getattr(self, "_cell_gradient_z", None) is None:
            nFx = self.nFx
            nFy = self.nFy
            i_s = self.face_boundary_indices

            iz = np.ones(self.nFz)
            iz[i_s[4]] = 0.0
            iz[i_s[5]] = 0.0
            Pafz = sp.diags(iz)

            MfI = self.get_face_inner_product(invert_matrix=True)
            MfIz = sp.diags(MfI.diagonal()[nFx + nFy :])

            self._cell_gradient_z = (
                -Pafz * MfIz * self.face_z_divergence.T * sp.diags(self.cell_volumes)
            )

        return self._cell_gradient_z

    @property
    def face_x_divergence(self):  # NOQA D102
        # Documentation inherited from discretize.operators.DifferentialOperators
        if getattr(self, "_face_x_divergence", None) is None:
            self._face_x_divergence = self.face_divergence[:, : self.nFx]
        return self._face_x_divergence

    @property
    def face_y_divergence(self):  # NOQA D102
        # Documentation inherited from discretize.operators.DifferentialOperators
        if getattr(self, "_face_y_divergence", None) is None:
            self._face_y_divergence = self.face_divergence[
                :, self.nFx : self.nFx + self.nFy
            ]
        return self._face_y_divergence

    @property
    def face_z_divergence(self):  # NOQA D102
        # Documentation inherited from discretize.operators.DifferentialOperators
        if getattr(self, "_face_z_divergence", None) is None:
            self._face_z_divergence = self.face_divergence[:, self.nFx + self.nFy :]
        return self._face_z_divergence

    def point2index(self, locs):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        return self.get_containing_cell_indexes(locs)

    def cell_levels_by_index(self, indices):
        """Fast function to return a list of levels for the given cell indices.

        Parameters
        ----------
        indices: (N) array_like
            Cell indexes to query

        Returns
        -------
        (N) numpy.ndarray of int
            Levels for the cells.
        """
        return self._cell_levels_by_indexes(indices)

    def get_interpolation_matrix(  # NOQA D102
        self, locs, location_type="cell_centers", zeros_outside=False, **kwargs
    ):
        # Documentation inherited from discretize.base.BaseMesh
        if "locType" in kwargs:
            raise TypeError(
                "The locType keyword argument has been removed, please use location_type. "
                "This will be removed in discretize 1.0.0"
            )
        if "zerosOutside" in kwargs:
            raise TypeError(
                "The zerosOutside keyword argument has been removed, please use zeros_outside. "
                "This will be removed in discretize 1.0.0"
            )
        location_type = self._parse_location_type(location_type)
        if self.dim == 2 and "z" in location_type:
            raise NotImplementedError("Unable to interpolate from Z edges/faces in 2D")

        locs = self._require_ndarray_with_dim("locs", locs, ndim=2, dtype=np.float64)
        if location_type == "nodes":
            Av = self._getNodeIntMat(locs, zeros_outside)
        elif location_type in ["edges_x", "edges_y", "edges_z"]:
            Av = self._getEdgeIntMat(locs, zeros_outside, location_type[-1])
        elif location_type in ["faces_x", "faces_y", "faces_z"]:
            Av = self._getFaceIntMat(locs, zeros_outside, location_type[-1])
        elif location_type in ["cell_centers"]:
            Av = self._getCellIntMat(locs, zeros_outside)
        else:
            raise ValueError(
                "Location must be a grid location, not {}".format(location_type)
            )
        return Av

    @property
    def permute_cells(self):
        """Permutation matrix re-ordering of cells sorted by x, then y, then z.

        Returns
        -------
        (n_cells, n_cells) scipy.sparse.csr_matrix
        """
        # TODO: cache these?
        P = np.lexsort(self.gridCC.T)  # sort by x, then y, then z
        return sp.identity(self.nC).tocsr()[P]

    @property
    def permute_faces(self):
        """Permutation matrix re-ordering of faces sorted by x, then y, then z.

        Returns
        -------
        (n_faces, n_faces) scipy.sparse.csr_matrix
        """
        # TODO: cache these?
        Px = np.lexsort(self.gridFx.T)
        Py = np.lexsort(self.gridFy.T) + self.nFx
        if self.dim == 2:
            P = np.r_[Px, Py]
        else:
            Pz = np.lexsort(self.gridFz.T) + (self.nFx + self.nFy)
            P = np.r_[Px, Py, Pz]
        return sp.identity(self.nF).tocsr()[P]

    @property
    def permute_edges(self):
        """Permutation matrix re-ordering of edges sorted by x, then y, then z.

        Returns
        -------
        (n_edges, n_edges) scipy.sparse.csr_matrix
        """
        # TODO: cache these?
        Px = np.lexsort(self.gridEx.T)
        Py = np.lexsort(self.gridEy.T) + self.nEx
        if self.dim == 2:
            P = np.r_[Px, Py]
        if self.dim == 3:
            Pz = np.lexsort(self.gridEz.T) + (self.nEx + self.nEy)
            P = np.r_[Px, Py, Pz]
        return sp.identity(self.nE).tocsr()[P]

    @property
    def cell_state(self):
        """The current state of the cells on the mesh.

        This represents the x, y, z indices of the cells in the base tensor mesh, as
        well as their levels. It can be used to reconstruct the mesh.

        Returns
        -------
        dict
            dictionary with two entries:

            - ``"indexes"``: the indexes of the cells
            - ``"levels"``: the levels of the cells
        """
        indexes, levels = self.__getstate__()
        return {"indexes": indexes.tolist(), "levels": levels.tolist()}

    def validate(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        return self.finalized

    def equals(self, other):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        try:
            if self.finalized and other.finalized:
                return super().equals(other)
        except AttributeError:
            pass
        return False

    def __reduce__(self):
        """Return the necessary items to reconstruct this object's state."""
        return TreeMesh, (self.h, self.origin), self.__getstate__()

    cellGrad = deprecate_property(
        "cell_gradient", "cellGrad", removal_version="1.0.0", error=True
    )
    cellGradx = deprecate_property(
        "cell_gradient_x", "cellGradx", removal_version="1.0.0", error=True
    )
    cellGrady = deprecate_property(
        "cell_gradient_y", "cellGrady", removal_version="1.0.0", error=True
    )
    cellGradz = deprecate_property(
        "cell_gradient_z", "cellGradz", removal_version="1.0.0", error=True
    )
    cellGradStencil = deprecate_property(
        "cell_gradient_stencil",
        "cellGradStencil",
        removal_version="1.0.0",
        error=True,
    )
    faceDivx = deprecate_property(
        "face_x_divergence", "faceDivx", removal_version="1.0.0", error=True
    )
    faceDivy = deprecate_property(
        "face_y_divergence", "faceDivy", removal_version="1.0.0", error=True
    )
    faceDivz = deprecate_property(
        "face_z_divergence", "faceDivz", removal_version="1.0.0", error=True
    )
    maxLevel = deprecate_property(
        "max_used_level", "maxLevel", removal_version="1.0.0", error=True
    )
    areaFx = deprecate_property(
        "face_x_areas", "areaFx", removal_version="1.0.0", error=True
    )
    areaFy = deprecate_property(
        "face_y_areas", "areaFy", removal_version="1.0.0", error=True
    )
    areaFz = deprecate_property(
        "face_z_areas", "areaFz", removal_version="1.0.0", error=True
    )
    edgeEx = deprecate_property(
        "edge_x_lengths", "edgeEx", removal_version="1.0.0", error=True
    )
    edgeEy = deprecate_property(
        "edge_y_lengths", "edgeEy", removal_version="1.0.0", error=True
    )
    edgeEz = deprecate_property(
        "edge_z_lengths", "edgeEz", removal_version="1.0.0", error=True
    )
    permuteCC = deprecate_property(
        "permute_cells", "permuteCC", removal_version="1.0.0", error=True
    )
    permuteF = deprecate_property(
        "permute_faces", "permuteF", removal_version="1.0.0", error=True
    )
    permuteE = deprecate_property(
        "permute_edges", "permuteE", removal_version="1.0.0", error=True
    )
    faceBoundaryInd = deprecate_property(
        "face_boundary_indices",
        "faceBoundaryInd",
        removal_version="1.0.0",
        error=True,
    )
    cellBoundaryInd = deprecate_property(
        "cell_boundary_indices",
        "cellBoundaryInd",
        removal_version="1.0.0",
        error=True,
    )
    _aveCC2FxStencil = deprecate_property(
        "average_cell_to_total_face_x",
        "_aveCC2FxStencil",
        removal_version="1.0.0",
        error=True,
    )
    _aveCC2FyStencil = deprecate_property(
        "average_cell_to_total_face_y",
        "_aveCC2FyStencil",
        removal_version="1.0.0",
        error=True,
    )
    _aveCC2FzStencil = deprecate_property(
        "average_cell_to_total_face_z",
        "_aveCC2FzStencil",
        removal_version="1.0.0",
        error=True,
    )
    _cellGradStencil = deprecate_property(
        "stencil_cell_gradient",
        "_cellGradStencil",
        removal_version="1.0.0",
        error=True,
    )
    _cellGradxStencil = deprecate_property(
        "stencil_cell_gradient_x",
        "_cellGradxStencil",
        removal_version="1.0.0",
        error=True,
    )
    _cellGradyStencil = deprecate_property(
        "stencil_cell_gradient_y",
        "_cellGradyStencil",
        removal_version="1.0.0",
        error=True,
    )
    _cellGradzStencil = deprecate_property(
        "stencil_cell_gradient_z",
        "_cellGradzStencil",
        removal_version="1.0.0",
        error=True,
    )



from libcpp cimport bool

cdef extern from "geom.h":
    ctypedef int int_t
    cdef cppclass Ball:
        Ball() except +
        Ball(int_t dim, double * x0, double r) except +

    cdef cppclass Line:
        Line() except +
        Line(int_t dim, double * x0, double *x1) except +

    cdef cppclass Box:
        Box() except +
        Box(int_t dim, double * x0, double *x1) except +

    cdef cppclass Plane:
        Plane() except +
        Plane(int_t dim, double * origin, double *normal) except +

    cdef cppclass Triangle:
        Triangle() except +
        Triangle(int_t dim, double * x0, double *x1, double *x2) except +

    cdef cppclass VerticalTriangularPrism:
        VerticalTriangularPrism() except +
        VerticalTriangularPrism(int_t dim, double * x0, double *x1, double *x2, double h) except +

    cdef cppclass Tetrahedron:
        Tetrahedron() except +
        Tetrahedron(int_t dim, double * x0, double *x1, double *x2, double *x3) except +


"""Deprecated view module."""

from discretize.utils.code_utils import deprecate_module

deprecate_module(
    "discretize.View",
    "discretize.mixins.mpl_mod",
    removal_version="1.0.0",
    error=True,
)
try:
    from discretize.mixins.mpl_mod import Slicer  # NOQA F401
except ImportError:
    pass



"""
=====================================
Discretize Meshes (:mod:`discretize`)
=====================================
.. currentmodule:: discretize

The ``discretize`` package contains four types of meshes for soliving partial differential
equations using the finite volume method.

Mesh Classes
============
.. autosummary::
  :toctree: generated/

  TensorMesh
  CylindricalMesh
  CurvilinearMesh
  TreeMesh
  SimplexMesh

Mesh Cells
==========
The :class:`~discretize.tensor_cell.TensorCell` and
:class:`~discretize.tree_mesh.TreeCell` classes were designed specifically to
define the cells within tensor and tree meshes, respectively.
Instances of :class:`~discretize.tree_mesh.TreeCell` and
:class:`~discretize.tensor_cell.TensorCell` are not meant to be created on
their own.
However, they can be returned directly by indexing a particular cell within
a tensor or tree mesh.

.. autosummary::
  :toctree: generated/

  tensor_cell.TensorCell
  tree_mesh.TreeCell
"""

from discretize.tensor_mesh import TensorMesh
from discretize.cylindrical_mesh import CylMesh, CylindricalMesh
from discretize.curvilinear_mesh import CurvilinearMesh
from discretize.unstructured_mesh import SimplexMesh
from discretize.utils.io_utils import load_mesh
from .tensor_cell import TensorCell

try:
    from discretize.tree_mesh import TreeMesh
except ImportError as err:
    import os

    # Check if being called from non-standard location (i.e. a git repository)
    # is tree_ext.pyx here? will not be in the folder if installed to site-packages...
    file_test = os.path.dirname(os.path.abspath(__file__)) + "/_extensions/tree_ext.pyx"
    if os.path.isfile(file_test):
        # Then we are being run from a repository
        raise ImportError(
            """
            It would appear that discretize is being imported from its source code
            directory and is unable to load its compiled extension modules. Try changing
            your directory and re-launching your python interpreter.

            If this was intentional, you need to install discretize in an editable mode.
            """
        )
    else:
        raise err
from discretize import tests

__author__ = "SimPEG Team"
__license__ = "MIT"
__copyright__ = "2013 - 2023, SimPEG Developers, https://simpeg.xyz"

from importlib.metadata import version, PackageNotFoundError

# Version
try:
    # - Released versions just tags:       0.8.0
    # - GitHub commits add .dev#+hash:     0.8.1.dev4+g2785721
    # - Uncommitted changes add timestamp: 0.8.1.dev4+g2785721.d20191022
    __version__ = version("discretize")
except PackageNotFoundError:
    # If it was not installed, then we don't know the version. We could throw a
    # warning here, but this case *should* be rare. discretize should be
    # installed properly!
    from datetime import datetime

    __version__ = "unknown-" + datetime.today().strftime("%Y%m%d")
